"""File with the for the SXPSQuery class."""

__docformat__ = "restructedtext en"

import pandas as pd
from .dataquery_base import dataQuery
from ..data import SXPS as dcat
from .. import plotLightCurve as lcPlot


class SXPSQuery(dataQuery):
    """A request to query the SXPS catalogues.

    This inherits from the dataQuery class, providing access to the
    SXPS catalogues; 2SXPS and LSXPS.

    This adds additional functionality, for selecting which SXPS
    catalogue to query and for obtaining more data than is available
    from a simple catalogue query.

    """

    def __init__(self, cat="LSXPS", table=None, silent=True, verbose=False):
        super().__init__(silent=silent, verbose=verbose)
        """Create an SXPSQuery instance.

        Parameters
        ----------

        cat : str, optional
            Which SXPS catalogue to base this query on
            (default: 'LSXPS').

        table : str or ``None``
            Which table to base the query on. If ``None``, uses the
            catalogue default.

        silent : bool
            Whether to suppress all console output (default: ``True``).

        verbose : bool
            Whether to give verbose output for everything
            (default: False; overridden by ``silent``).

        """

        # This constructor is slightly more involved than the basic, because
        # we support multiple catalogues = multiple datasets.
        # This means that I need to list the supported catalogues, and which
        # tables they support.
        # This can be done all in one, by a dict of tablesByCat, where the keys
        # are the catalogues and the
        self._tablesByCat = {
            "LSXPS": [
                "sources",
                "datasets",
                "detections",
                "obssources",
                "xcorr",
                "transients",
                "oldstacks",
            ],
            "2SXPS": [
                "sources",
                "datasets",
                "detections",
                "xcorr",
            ],
        }
        # Also need the default table to select when a catalogue is selected.
        self._defaultTablesByCat = {
            "LSXPS": "sources",
            "2SXPS": "sources",
        }

        # Default ID col for astroquery lookups
        self._defaultIDColBySource = {
            "LSXPS": {
                "sources": "LSXPS_ID",
                "datasets": "Dataset_ID",
                "transients": "TransientID",
            },
            "2SXPS": {"sources": "2SXPS_ID", "datasets": "ObsID"},
        }

        self._subsets = {
            "LSXPS": {
                "sources": ("clean", "ultraclean"),
                "datasets": ("clean", "ultraclean"),
            },
            "2SXPS": {
                "sources": ("clean", "ultraclean"),
                "datasets": ("clean", "ultraclean"),
            },
        }

        # Now set the selected catalogue to that supplied in the constructor.
        # As cat is actually a property dectorator, this will check that
        # the catalogue is valid and sert the tables and default table.
        self.cat = cat
        self._subset = None
        if table is not None:
            self.table = table

        self._prodData.update(
            {
                "sourceDetails": None,
                "lightCurves": None,
                "spectra": None,
                "datasetDetails": None,
                "transientDetails": None,
                "fullTable": None,
                "sourceObsList": None,
            }
        )

        self._lcbinning = None
        self._lctimeformat = None

    # ----------------------------------------------------------
    # Some properties unique to this class.
    @property
    def cat(self):
        """Which SXPS catalogue is to be queried."""
        return self._dbName

    @cat.setter
    def cat(self, dbName):
        if dbName in self.cats:
            self._dbName = dbName
            self._tables = self._tablesByCat[dbName]
            self.table = self._defaultTablesByCat[dbName]
        else:
            raise ValueError(f"{dbName} is not a known database.")

    @property
    def cats(self):
        """List of queryable SXPS catalogues."""
        return list(self._tablesByCat.keys())

    @property
    def sourceDetails(self):
        """Detailed source information."""
        return self._prodData["sourceDetails"]

    @property
    def transientDetails(self):
        """Detailed transient information."""
        return self._prodData["transientDetails"]

    @property
    def lightCurves(self):
        """Source light curves."""
        return self._prodData["lightCurves"]

    @property
    def spectra(self):
        """Source spectra."""
        return self._prodData["spectra"]

    @property
    def datasetDetails(self):
        """Detailed dataset information."""
        return self._prodData["datasetDetails"]

    @property
    def fullTable(self):
        """The full catalogue."""
        return self._prodData["fullTable"]

    @property
    def sourceObsList(self):
        """Detailed dataset information."""
        return self._prodData["sourceObsList"]

    @property
    def subsets(self):
        """The available subsets for the selected table"""
        return self._subsets[self._dbName][self.table]

    @property
    def subset(self):
        """The selected subset"""
        return self._subset

    @subset.setter
    def subset(self, set):
        if set is None:
            self._subset = None
            if self.verbose:
                print("Removing subset filters.")
        elif self._subsets is None:
            raise RuntimeError(f"{self.cat}.{self.table} has no subset.")
        elif set.lower() in self._subsets[self._dbName][self.table]:
            self._subset = set
            if self.verbose:
                print(f"Selecting `{set}` subset.")
        else:
            raise ValueError(f"`{set}` is not a valid subset.")

    # Also override the table setter property, so that we can accept
    # short-form table names and correct them.
    @dataQuery.table.setter
    def table(self, table):
        if table.lower() in dcat.tableLookup:
            table = dcat.tableLookup[table.lower()]
            if not self.silent:
                print(f"Adjusting table name to `{table}`")
        dataQuery.table.fset(self, table.lower())

        # Update default ID col if necessary
        if table in self._defaultIDColBySource[self._dbName]:
            self._defaultIDCol = self._defaultIDColBySource[self._dbName][table]
        else:
            self._defaultIDCol = None

        self.subset = None

    # Work out which new fns or functionality must be added:
    # (Labels on these are not final):
    #
    # getDetails() - for sources/datasets essentially gets all the web page info
    # Something to get products; images, LC, spectra &c
    #   Both of the above need to be able to run for a set of results
    #
    # setSubset() - need to be able to choose clean/ultraclean
    #           EXTRA NOTE: Ideally handle this at the back end, so that
    #             the subset definitions are only defined once. Have to
    #             then incorporate that into website.
    #             IDEA: Function to "get subset" which creates and returns a filter object/list?
    #
    # Upper limit generator.

    # ------------------------------------------------
    # Need to override the submit function.
    def submit(self):
        """Submit the query.

        This checks if a subset needs adding before calling the parent
        submit() function.

        """
        if self.subset is None:
            return super().submit()
        else:
            return super().submit(subset=self.subset)

    # And reset
    def reset(self, **kwargs):
        super().reset(**kwargs)
        self.clearLightCurves()

    # ------------------------------------------------
    # Functions to retrieve information.

    def _handleSourceArgs(self, byName=False, byID=False):
        """Decide whether to query on sourceID or name.

        This is an internal helper function for any function that gets
        products related to a query. It decides whether to send the
        sourceName or ID column with the request, depending first,
        on whether either was explicitly requested, and if not, based
        on which columns have been retrieved.

        Parameters
        ----------

        byName : bool
            Whether to force use of the source name.

        byID : bool
            Whether to force use of the source ID

        Returns
        -------
        whichCol : str
            The name o the column to index on in the returned query.
        whichArg : str
            The name of the argument to send to the API.

        """
        whichCol = None
        whichArg = None

        if byName:
            whichCol = "IAUName"
            whichArg = "sourceName"
            if byID:
                raise RuntimeError("You cannot set byName and byID!")
        elif byID:
            whichCol = self._defaultIDCol  # f"{self.cat}_ID"
            whichArg = "sourceID"
        else:
            cols = {f"{self._defaultIDCol}": "sourceID", "IAUName": "sourceName"}

            for c, n in cols.items():
                if c in self.results.columns:
                    whichCol = c
                    whichArg = n
                    break
            if whichCol is None:
                raise RuntimeError(f"Cannot get source info, as none of the columns: {cols} are in your results.")

        if whichCol not in self.results.columns:
            raise RuntimeError(f"Cannot get source info, as the column `{whichCol}` is not in your results.")

        return (whichCol, whichArg)

    def _handleDatasetArgs(self, byObsID=False, byDatasetID=False):
        """Decide whether to query on sourceID or name.

        This is an internal helper function for any function that gets
        products related to a query. It decides whether to send the
        DatasetID or obsID column with the request, depending first,
        on whether either was explicitly requested, and if not, based
        on which columns have been retrieved.

        Parameters
        ----------

        byObsID : bool
            Whether to force use of the ObsID.

        byDatasetID : bool
            Whether to force use of the DatasetID

        Returns
        -------
        whichCol : str
            The name o the column to index on in the returned query.
        whichArg : str
            The name of the argument to send to the API.

        """
        whichCol = None
        whichArg = None

        if byObsID:
            whichCol = "ObsID"
            whichArg = "ObsID"
            if byDatasetID:
                raise RuntimeError("You cannot set byName and byID!")
        elif byDatasetID:
            whichCol = "DatasetID"
            whichArg = "DatasetID"
        else:
            cols = ("ObsID", "DatasetID")

            for c in cols:
                if c in self.results.columns:
                    whichCol = c
                    whichArg = c
                    break
            if whichCol is None:
                raise RuntimeError(f"Cannot get dataset info, as none of the columns: {cols} are in your results.")

        if whichCol not in self.results.columns:
            raise RuntimeError(f"Cannot get dataset info, as the column `{whichCol}` is not in your results.")

        return (whichCol, whichArg)

    def getSourceDetails(self, subset=None, byName=False, byID=False, returnData=False):
        """Get the full set of information for retrieved sources.

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.getSourceDetails; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        The retrieved information will be saved in the "sourceDetails"
        variable of this object; it will also be returned by this
        function call if returnedData=True

        Left alone, this will get the details for all sources your query
        identified, and index them by SXPS ID (if that was retrieved),
        or IAUName if the ID was not retrieved. If neither column was
        retrieved, expect an error.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getSourceDetails(subset=q.results['Err90']<5)

        would get details of those sources in your result with Err90
        values <5.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether to return the data, as well as storing it in the
            "sourceDetails" variable.

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        if self.table != "sources":
            return RuntimeError("Cannot get source info for anything except sources!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        tmp = dcat.getSourceDetails(cat=self.cat, silent=self.silent, verbose=self.verbose, **data)

        if self._prodData["sourceDetails"] is None:
            self._prodData["sourceDetails"] = tmp
        else:
            self._prodData["sourceDetails"].update(tmp)

        if not self.silent:
            print("Saved source information as sourceDetails varable.")
        if returnData:
            return self._prodData["sourceDetails"]

    def getDatasetDetails(self, byObsID=False, byDatasetID=False, subset=None, returnData=False):
        """Get the full set of information for retrieved datasets.

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.getDatasetDetails; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        The retrieved information will be saved in the "datasetDetails"
        variable of this object; it will also be returned by this
        function call if returnedData=True

        Left alone, this will get the details for all datasets supplied,
        identified, and index them by DatasetID (if that was retrieved),
        or ObsID otherwise. If neither column was retrieved, expect an
        error.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getDatasetDetails(subset=q.results['ExposureUsed']>10000)

        would get details of all those datasets in your result with
        exposures longer than 10ks.

        Parameters
        ----------

        byObsID : bool, optional
            Index the results by ObsID (default: False).

        byDatasetID : bool, optional
            Index the results by DatasetID (default: False).

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        returnData : bool, optional
            Whether to return the data, as well as storing it in the
            "datasetDetails" variable.

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        if self.table != "datasets":
            return RuntimeError("Cannot get dataset info for anything other than datasets!")

        # Check if we are doing ID or name.
        data = {}

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        (whichCol, whichArg) = self._handleDatasetArgs(byObsID=byObsID, byDatasetID=byDatasetID)

        obslist = []
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            obslist = self._results.loc[subset][whichCol].tolist()
        else:
            obslist = self.results[whichCol].tolist()

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichCol] = obslist

        tmp = dcat.getDatasetDetails(cat=self.cat, silent=self.silent, verbose=self.verbose, **data)

        if self._prodData["datasetDetails"] is None:
            self._prodData["datasetDetails"] = tmp
        else:
            self._prodData["datasetDetails"].update(tmp)

        if not self.silent:
            print("Saved dataset information as datasetDetails varable.")
        if returnData:
            return self._prodData["datasetDetails"]

    def getTransientDetails(self, subset=None, byName=False, byID=False, returnData=False):
        """Get the full set of information for retrieved transients.

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.getTransientDetails; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        The retrieved information will be saved in the "transientDetails"
        variable of this object; it will also be returned by this
        function call if returnedData=True

        Left alone, this will get the details for all sources your query
        identified, and index them by TransientID (if that was retrieved),
        or IAUName if the ID was not retrieved. If neither column was
        retrieved, expect an error.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getTransientDetails(subset=q.results['Err90']<5)

        would get details of those sources in your result with Err90
        values <5.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether to return the data, as well as storing it in the
            "transientDetails" variable.

        """
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        if self.table != "transients":
            raise RuntimeError("Cannot get transient info for anything except transients!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        self._prodData["transientDetails"] = dcat.getTransientDetails(
            cat=self.cat, silent=self.silent, verbose=self.verbose, **data
        )
        if not self.silent:
            print("Saved transient information as transientDetails varable.")
        if returnData:
            return self._prodData["transientDetails"]

    def getDetails(self, **kwargs):
        """A wrapper to call the appropriate function."""
        if self.table == "sources":
            self.getSourceDetails(**kwargs)
        elif self.table == "datasets":
            self.getDatasetDetails(**kwargs)
        elif self.table == "transients":
            self.getTransientDetails(**kwargs)
        else:
            raise NotImplementedError

    def getFullTable(self, returnData=False, **kwargs):
        """Get the full catalogue data.

        This downloads the entire table in one go, storing it in the
        ``fullTable`` variable. You can optionally supply a subset, for
        the tables which have defined subset.

        This wraps swifttools.ukssdc.data.SXPS.getFullTable, and the
        ``**kwargs`` are passed to that function.

        Parameters
        ----------

        subset : str, optional
            Some tables have subsets you can select, e.g. the 'clean'
            set of sources. You can specify that subset using
            this parameter or, if this is not supplied, the default
            subset will be obtained.

        returnData : bool, optional
            Whether to return the data, as well as storing it in the
            ``fullTable`` variable.

        """

        if "saveData" not in kwargs:
            kwargs["saveData"] = False

        self._prodData["fullTable"] = dcat.getFullTable(
            cat=self.cat,
            table=self.table,
            subset=self.subset,
            returnData=True,
            silent=self.silent,
            verbose=self.verbose,
            **kwargs,
        )
        if returnData:
            return self.fullTable

    # ------------------------------------------------
    # Functions to retrieve products

    def getLightCurves(
        self,
        subset=None,
        byName=False,
        byID=False,
        bands="all",
        binning=None,
        timeFormat=None,
        returnData=False,
        **kwargs,
    ):
        """Get the full light curves for retrieved sources.

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.getLightCurves; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        Left alone, this will get the details for all sources your query
        identified, and index them by SXPS ID (if that was retrieved),
        or IAUName if the ID was not retrieved. If neither column was
        retrieved, expect an error.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getLightCurves(subset=q.results['Err90']<5)

        would get lightcurves of all of the sources in your result with
        Err90 values <5.


        The light curves returned are stored in the ``lightCurves``
        variable of this object, and optionally returned. For their
        format see the help for
        swifttools.ukssdc.data.SXPS.getLightCurves.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        bands : str or list or tuple
            Either 'all', or a list of labels of the energy bands to
            retrieve ('total', 'soft', 'medium', 'hard', 'HR1', 'HR2')

        binning : str
            Which of the two binning methods you want: either
            'observation' or 'snapshot'

        timeFormat : str
            Which units to use for the time axis. Must be 'met' or 'tdb'
            or 'mjd'.

        returnData : bool, optional
            Whether the light curve data should be returned by this
            function, as well as saved in the "lightCurves"
            variable.

        **kwargs : dict
            Other arguments which are passed to
            swifttools.ukssdc.data.SXPS.getlightCurves.
            See its help for more information.

        """
        if self.table not in ("sources", "transients"):
            raise RuntimeError("Cannot get light curves for anything other than sources or transients!")

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        # If we already have light curves, then we have to make sure
        # that the binning and timeformat match. If none were supplied then
        # we will set them to what was got before.
        if binning is not None:
            binning = binning.lower()
            if binning not in dcat.LC_BINNING:
                raise ValueError(f"{binning} is not an acceptable binning method.")
            binning = dcat.LC_BINNING[binning]

        if timeFormat is not None:
            timeFormat = timeFormat.lower()
            if timeFormat not in dcat.LC_TIMEFORMAT:
                raise ValueError(f"{timeFormat} is not an acceptable time system.")
            timeFormat = dcat.LC_TIMEFORMAT[timeFormat]

        if self._prodData["lightCurves"] is not None:
            if binning is None:
                binning = self._lcbinning
            if timeFormat is None:
                timeFormat = self._lctimeformat
            if (binning != self._lcbinning) or (timeFormat != self._lctimeformat):
                raise ValueError("Cannot get more light curves with different binning / timeformat.")

        useTrans = False
        if self.table == "transients":
            useTrans = True

        if "saveData" not in kwargs:
            kwargs["saveData"] = False

        tmp = dcat.getLightCurves(
            cat=self.cat,
            silent=self.silent,
            verbose=self.verbose,
            bands=bands,
            binning=binning,
            transient=useTrans,
            timeFormat=timeFormat,
            returnData=True,
            **data,
            **kwargs,
        )

        if self._prodData["lightCurves"] is None:
            self._prodData["lightCurves"] = tmp
        else:
            # Can't just merge the dicts, annoyingly, because this is not a deep (recursive) merge, so will not
            # add extra bands. However, I don't need a full recurse, just at the source level, so it's not too hard.
            for i in tmp.keys():
                if i in self._prodData["lightCurves"]:
                    tmpDS = self._prodData["lightCurves"][i]["Datasets"]
                    tmpDS.extend(x for x in tmp[i]["Datasets"] if x not in tmpDS)
                    self._prodData["lightCurves"][i].update(tmp[i])
                    # self._prodData['lightCurves'][i] = {**self._prodData['lightCurves'][i], **tmp[i]}
                    self._prodData["lightCurves"][i]["Datasets"] = tmpDS
                else:
                    self._prodData["lightCurves"][i] = tmp[i]

        # Now get the binning and timeformat, assuming any exist
        tmp = list(self._prodData["lightCurves"].keys())
        firstLC = tmp[0]
        if "Binning" in self._prodData["lightCurves"][firstLC]:
            self._lcbinning = self._prodData["lightCurves"][firstLC]["Binning"]
        if "TimeFormat" in self._prodData["lightCurves"][firstLC]:
            self._lctimeformat = self._prodData["lightCurves"][firstLC]["TimeFormat"]

        if returnData:
            return self._prodData["lightCurves"]

    def saveLightCurves(self, **kwargs):
        """Save the light curves to text files.

        This function is used to save the light curves that have been
        downloaded and are stored in the self.lightCurves
        variable of this SXPSQuery object. These light curves
        can be saved as simple text file (with a user-settable
        column separator), or as files formatted for the qdp programme.

        All this function actually does it to call the
        saveLightCurves() in the related module:
        swifttools.ukssdc.data; further detail on the parameters that
        can be used are documented for that function.

        Parameters
        ----------

        **kwargs : dict
            A set of parameters to pass to the main save function in
            swifttools.ukssdc.data

        """
        # Check that we have downloaded some light curves!
        if self.lightCurves is None:
            raise RuntimeError("There are no light curves to save!")
        dcat.saveLightCurves(self.lightCurves, silent=self.silent, verbose=self.verbose, **kwargs)

    def clearLightCurves(self):
        """Forget all light curves."""
        self._prodData["lightCurves"] = None
        self._lcbinning = None
        self._lctimeformat = None

    def plotLightCurves(self, whichSource, **kwargs):
        """Plot the light curves of a specific GRB.

        This is a wrapper to the ``ukssdc.plotLightCurve()`` function,
        giving you a plot of the light curve in question.

        You can only plot one GRB at a time (at present) and so you
        must provide the identifier (name or targetID) of a single
        downloaded GRB.

        Parameters
        ----------

        whichSource : str
            The identifier of a GRB to download. Must be a key to the
            ``lightCurves`` dict.

        **kwargs : dict
            Arguments to pass to ``plotLightCurve()``

        """
        if self.lightCurves is None:
            raise RuntimeError("I don't have any light curves to plot!")

        if whichSource not in self.lightCurves:
            raise RuntimeError(f"`{whichSource}` is not a recognised, downloaded GRB.")

        if "silent" not in kwargs:
            kwargs["silent"] = self.silent
        if "verbose" not in kwargs:
            kwargs["verbose"] = self.verbose

        return lcPlot(self.lightCurves[whichSource], **kwargs)

    def getSpectra(self, subset=None, byName=False, byID=False, returnData=False, **kwargs):
        """Get the spectral information for the retrieved source(s).

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.getSpectra; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        Left alone, this will get the details for all sources your query
        identified, and index them by SXPS ID (if that was retrieved),
        or IAUName if the ID was not retrieved. If neither column was
        retrieved, expect an error.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getSpectra(subset=q.results['Err90']<5)

        would get spectra for the sources in your result with Err90
        values <5.

        The light curves returned are stored in the ``spectra``
        variable of this object, and optionally returned. For their
        format see the help for swifttools.ukssdc.data.SXPS.getSpectra.

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        returnData : bool, optional
            Whether the light curve data should be returned by this
            function, as well as saved in the "lightCurves"
            variable.

        **kwargs : dict, optional
            Any arguments to pass to ``data.SXPS.getSpectra``, e.g.
            specType if this is a transient.

        """
        if self.table not in ("sources", "transients"):
            raise RuntimeError("Cannot get spectra for anything other than sources or transients!")

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        useTrans = False
        if self.table == "transients":
            useTrans = True

        if "saveData" not in kwargs:
            kwargs["saveData"] = False

        tmp = dcat.getSpectra(
            cat=self.cat,
            silent=self.silent,
            verbose=self.verbose,
            transient=useTrans,
            returnData=True,
            **data,
            **kwargs,
        )

        if self._prodData["spectra"] is None:
            self._prodData["spectra"] = tmp
        else:
            self._prodData["spectra"].update(tmp)

        if returnData:
            return self._prodData["spectra"]

    def saveSpectra(self, **kwargs):
        """Save the spectra files to disk.

        This function is used to save the spectra that have been
        downloaded and are stored in the self.spectra
        variable of this SXPSQuery object.

        All this function actually does it to call the
        saveSpectra() in the related module:
        swifttools.ukssdc.data; further detail on the parameters that
        can be used are documented for that function.

        Parameters
        ----------

        **kwargs : dict
            A set of parameters to pass to the main save function in
            swifttools.ukssdc.data

        """
        # Check that we have downloaded some light curves!
        if self.spectra is None:
            raise RuntimeError("There are no spectra to save!")
        dcat.saveSpectra(self.spectra, silent=self.silent, verbose=self.verbose, **kwargs)

    def clearSpectra(self):
        """Forget all spectra."""
        self._prodData["spectra"] = None

    def saveSourceImages(self, byName=False, byID=False, subset=None, **kwargs):
        """Download and save the images for retrieved datasets.

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.saveSoureImages; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        Left alone, this will get the images for all sources your query
        identified, and save them all into a single directory. If you
        set ``subDirs=True`` then instead the images will be saved into
        one directory per source, with directories named by sourceID
        or name, depending on your byID/byName values.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getDatasetDetails(subset=q.results['ExposureUsed']>10000)

        would get images for all those datasets in your result with
        exposures longer than 10ks.

        Parameters
        ----------

        byName : bool, optional
            Index the results by sourceName (default: False).

        byID : bool, optional
            Index the results by sourceName (default: False).


        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        **kwargs : dict
            Arguments to pass to
            swifttools.ukssdc.data.SXPS.saveDatasetImages.

        """

        if self.table not in ("sources", "transients"):
            raise RuntimeError("Cannot get source images for anything other than sources or transients!")

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        data[whichArg] = []
        if self.table == "transients":
            data["transient"] = True

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        dcat.saveSourceImages(cat=self.cat, silent=self.silent, verbose=self.verbose, **data, **kwargs)

    def saveTransientImages(self, **kwargs):
        """Download and save the images for retrieved transients.

        Literally just redirects to saveSourceImages!
        """
        return self.saveSourceImages(**kwargs)

    def saveDatasetImages(self, byObsID=False, byDatasetID=False, subset=None, **kwargs):
        """Download and save the images for retrieved datasets.

        This function is primarily a wrapper to the function
        swifttools.ukssdc.data.SXPS.saveDatasetImages; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        Left alone, this will get the images for all datasets your query
        identified, and save them all into a single directory. If you
        set ``subDirs=True`` then instead the images will be saved into
        one directory per dataset, with directories named by DatasetID.
        If you set ``byObsID=True`` then the subdirectories will instead
        be the ObsIDs.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getDatasetDetails(subset=q.results['ExposureUsed']>10000)

        would get images for all those datasets in your result with
        exposures longer than 10ks.

        Parameters
        ----------

        byObsID : bool, optional
            Index the results by ObsID (default: False).

        byDatasetID : bool, optional
            Index the results by DatasetID (default: False).

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        **kwargs : dict
            Arguments to pass to
            swifttools.ukssdc.data.SXPS.saveDatasetImages.

        """

        if self.table != "datasets":
            raise RuntimeError("Cannot get dataset images for anything other than datasets!")

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        (whichCol, whichArg) = self._handleDatasetArgs(byObsID=byObsID, byDatasetID=byDatasetID)

        obslist = []
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            obslist = self._results.loc[subset][whichCol].tolist()
        else:
            obslist = self.results[whichCol].tolist()

        # Now set up data, which will will pass as **data which will be received either as
        # sourceID = [...] or sourceName = [...]
        data[whichCol] = obslist

        self._prodData["datasetDetails"] = dcat.saveDatasetImages(
            cat=self.cat, silent=self.silent, verbose=self.verbose, **data, **kwargs
        )

    def saveImages(self, **kwargs):
        """A wrapper to call the appropriate function."""
        if self.table == "sources":
            self.saveSourceImages(**kwargs)
        elif self.table == "datasets":
            self.saveDatasetImages(**kwargs)
        elif self.table == "transients":
            self.saveTransientImages(**kwargs)
        else:
            raise NotImplementedError

    # ------------------------------------------------
    # Interaction with xrt_prods
    def makeProductRequest(self, email, byName=False, byID=False, subset=None, **kwargs):
        """Create XRTProductRequest objects for your sources.

        This function is primarily a wrapper to the function
        ``swifttools.ukssdc.data.SXPS.makeProductRequest()``; calling
        that function for the source(s) which this query object has
        returned. For obvious reasons, it will fail if you have not yet
        executed the query.

        This will get the XRTProductRequests for all sources your query
        identified. For full details, see the help for
        swifttools.ukssdc.data.SXPS.makeProductRequest

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'q' then:

        => q.getDatasetDetails(subset=q.results['ExposureUsed']>10000)

        would get images for all those datasets in your result with
        exposures longer than 10ks.

        Parameters
        ----------

        email : str
            The email address with which you registered to submit
            XRTProductRequest objects to the servers.

        byName : bool, optional
            Index the results by ObsID, instead of the DatasetID
            (default: False).

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        **kwargs : dict
            Arguments to pass to
            swifttools.ukssdc.data.SXPS.saveDatasetImages.

        """
        if self.table not in ("sources", "transients"):
            raise RuntimeError("Cannot get XRTProductRequests for anything other than sources or transients!")
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        # Check if we are doing ID or name.
        data = {}

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        data[whichArg] = []
        if self.table == "transients":
            data["transient"] = True

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        return dcat.makeProductRequest(
            email,
            cat=self.cat,
            sourceDetails=self.sourceDetails,
            silent=self.silent,
            verbose=self.verbose,
            **data,
            **kwargs,
        )

    def getObsList(self, byName=False, byID=False, subset=None, returnData=False, **kwargs):
        """Get the list of observations covering a source.

        This function is primarily a wrapper to the function
        ``swifttools.ukssdc.data.SXPS.getObsList()``; calling
        that function for the source(s) which this query object has
        returned. For obvious reasons, it will fail if you have not yet
        executed the query.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        Parameters
        ----------

        byName : bool, optional
            Index the results by source name (default: ``False``).

        byID : bool, optional
            Index the results by source ID (default: ``False``).

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        returnData : bool, optional
            Whether to return the obslist found by this function call
            (default: ``False``).

        **kwargs : dict
            Arguments to pass to
            ``swifttools.ukssdc.data.SXPS.getObsList``.

        """
        if self.table != "sources":
            raise RuntimeError("Cannot get observation list for anything other than sources!")
        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot get extra data!")

        # Check if we are doing ID or name.
        data = {}

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        data[whichArg] = []

        # But before we can populate it, we may have to handle a subset:
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            data[whichArg] = self._results.loc[subset][whichCol].tolist()
        else:
            data[whichArg] = self._results[whichCol].tolist()

        tmp = dcat.getObsList(
            sourceDetails=self.sourceDetails,
            cat=self.cat,
            silent=self.silent,
            verbose=self.verbose,
            **data,
            **kwargs,
        )

        if self._prodData["sourceObsList"] is None:
            self._prodData["sourceObsList"] = tmp
        else:
            self._prodData["sourceObsList"].update(tmp)

        if returnData:
            return tmp

    # ----------------------------------------------------------------
    # astroquery-based functions

    def doSIMBADSearch(self, byName=False, byID=False, **kwargs):
        """Do a SIMBAD search around the object(s).

        This function is primarily a wrapper to the function
        swifttools.ukssdc.query.doSIMBADSearch; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        Parameters
        ----------

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        """
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        return super().doSIMBADSearch(**kwargs, idCol=whichCol)

    def doVizierSearch(self, byName=False, byID=False, **kwargs):
        """Do a Vizier search around the object(s).

        This function is primarily a wrapper to the function
        swifttools.ukssdc.query.doVizierSearch; calling that
        function for the source(s) which this query object has returned.
        For obvious reasons, it will fail if you have not yet executed
        the query.

        Parameters
        ----------

        byName : bool, optional
            Force the results to be indexed by IAUName.
            Requires that column to have been retrieved by your query.

        byID : bool, optional
            Force the results to be indexed by SXPS_ID.
            Requires that column to have been retrieved by your query.

        """
        (whichCol, whichArg) = self._handleSourceArgs(byName=byName, byID=byID)

        return super().doVizierSearch(**kwargs, idCol=whichCol)



"""ukssdc module, providing API interface to UKSSDC services.

This module provices an interface to various services offered by the
UK Swift Science Data Centre, which have traditionally been provided
via the website https://www.swift.ac.uk.

This README will be updated later.
"""

__all__ = ["APIURL"]

from .main import APIURL, plotLightCurve, bayesRate, mergeLightCurveBins, mergeUpperLimits  # noqa
from .version import __version__, _apiVersion  # noqa



__docformat__ = "restructedtext en"
from .dataquery_base import dataQuery


class ObsQuery(dataQuery):
    """A request to query the main Swift observation database.

    This inherits from the dataQuery class, and adds no new functions,
    it simply defines the database, tables and default columns for the
    observation database.

    """

    def __init__(self, table=None, silent=True, verbose=False):
        super().__init__(silent=silent, verbose=verbose)

        # Set the basic details for an obsquery - the database, acceptable
        # tables and default table.
        self._dbName = "obs"
        self._tables = ("swiftmastr", "swiftxrlog", "swiftbalog", "swiftuvlog", "swifttdrss")

        # Set the table, to default or to that requested.
        if table is not None:
            self.table = table
        else:
            self.table = "swiftmastr"



"""Module-wide variables and functions.

This file contains some constants and functions which are not intended
for export within the swiftools.ukssdc module, but which are common to
various things within that module. As such it should be considered to
be a 'private' module.

Essentially, this provides some core functionality which is of no use
in isolation but is called by the various classes.

Provided variables.
-------------------

_apiVersion : str  The version of the swifttools.ukssdc API.
    NOTE: This is distinct from the swifttools version, and the
    xrt_prods or swift_too versions; the various components of
    swifttools each have their own versions, and this is mainly used to
    warn users when there are back-end changes that should require the
    user to update their module.

_apiWarned : bool Whether a warning of an out-of-date module has already
    been issued this session.

APIURL : str The base URL directory for all API uploads.



Provided functions.
-------------------

submitAPICall() - actually submits an API request and checks the result
    is formatted correctly, i.e. the request was succesfully received
    and processed. Raises errors if there are problems.

"""

__docformat__ = "restructedtext en"


import requests
import json
import warnings
import math
import os
import numpy as np
import pandas as pd
from distutils.version import StrictVersion
from .version import _apiVersion


SXPS_BAND_NAMES = ("Total", "Soft", "Medium", "Hard")

HAS_ASTROPY = False
try:
    import astropy.coordinates

    HAS_ASTROPY = True
except ImportError:
    pass

HAS_ASTROQ = False
try:
    import astroquery  # noqa

    HAS_ASTROQ = True
except ImportError:
    pass


_apiWarned = False

APIURL = "https://www.swift.ac.uk/API/main.php"


# _funcList = {"getMetadata": "getMetadata", "queryDB": "queryDB", "listObs"}


def submitAPICall(func, data, minKeys=None, skipErrors=False, verbose=False):
    """Function to submit an API query and do simple validation.

    This function is designed for internal use by codes within the
    swifttools.ukssdc module; it is not intended for public use.

    This is a simple function to submit a specific query to our servers,
    and check that the returned JSON meets some minimum requirements. If
    the http return code is not 200, an 'ERROR' is returned, 'OK' is
    not returned, or a specified set of keys are not returned,
    this will raise an exception. This means that the different calls
    do not have to all check for success.

    Parameters
    ----------
        func : str
            The function to be carried out.

        data : dict
            A dictionary of the data to send as JSON with the request.

        skipErrors : bool, optional
            Whether to ignore any errors (default ``False``)

        minKeys: tuple, optonal
            A list of minimum keys that must be included in the return
            from the server. Default: None. NB, the "OK" key will ALWAYS
            be needed.

    Returns
    -------
        dict
            The set of returned data.

    """
    global _apiWarned

    # if func not in _funcList:
    #     raise ValueError(f"{func} is not a supported API function.")

    data["APIFunc"] = func
    data["APIVersion"] = _apiVersion

    if verbose:
        print(f"Uploading data to {APIURL}")
    #        print(data)

    sub = requests.post(APIURL, json=data)
    if sub.status_code != 200:
        print("Received HTTP failure from the server.")
        raise RuntimeError(f"An HTTP error occured - HTTP return code {sub.status_code}: {sub.reason}")

    # Pull the returned data into JSON.
    ret = json.loads(sub.text)

    # Check if we need to warn about the API
    if "APIVersion" in ret:
        if (StrictVersion(str(ret["APIVersion"])) > StrictVersion(_apiVersion)) and (not _apiWarned):
            warnings.warn(
                f"WARNING: you are using version {_apiVersion} of the UKSSDC API component; "
                f"the latest version is {ret['APIVersion']}, it would be advisable to update the swifttools module."
            )
            _apiWarned = True

    if verbose:
        print(f"Returned keys: {ret.keys()}")

    # Check for an error:
    if "ERROR" in ret and not skipErrors:
        msg = f"An error occurred processing your request. Code {ret['ERROR']}."
        if "ERRORTEXT" in ret:
            msg = msg + f" Message: `{ret['ERRORTEXT']}`"
        raise RuntimeError(msg)

    if "OK" not in ret and not skipErrors:
        raise RuntimeError("An unspecified error occured processing your request.")
    if minKeys is not None:
        if verbose:
            print("Checking returned data for required content.")
        bad = []
        for k in minKeys:
            if k not in ret:
                bad.append(k)
        if len(bad) > 0:
            msg = (
                "Several required properties were missing from the data returned by the server.\n"
                "This may mean that your swifttools version is out of date.\n"
                "If you have the latest version, please contact swift-help@leicester.ac.uk. The missing keys are\n"
            )
            msg = msg + ", ".join(bad) + "\n"
            raise RuntimeError(msg)

    # If we got here, then all is OK, so just return the dict we decoded from JSON
    # but first remove the keys we've handled.
    if not skipErrors:
        ret.pop("OK", None)
    ret.pop("APIVersion", None)

    return ret


def _createDir(destDir, silent=True, verbose=None):
    """Internal function to make a directory.

    Parameters
    ----------

    destDir : str
        Directory path

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    """
    if not os.path.isdir(destDir):
        if not silent:
            print(f"Making directory {destDir}")
        os.mkdir(destDir)
        if not os.path.isdir(destDir):
            raise RuntimeError(f"Cannot make directory {destDir}")


def makeAng(a):
    """Convert an angle into an astropy.coordinates.angle object."""
    if math.isnan(a):
        return None
    b = astropy.coordinates.Angle(a, unit="deg")
    return b


def makeSex(dec):
    """Convert angle from decimal to sexagesimal.

    Parameters
    ----------
    dec : float
        The decimal angle.

    Returns
    -------
    list
        The sexagesimal angle (d, m, s)

    """
    if math.isnan(dec):
        return None
    neg = False
    if dec < 0:
        neg = True

    dec = abs(dec)
    d = math.floor(dec)
    dec = (dec - d) * 60
    m = math.floor(dec)
    dec = (dec - m) * 60

    d = f"{d:02d}"
    if neg:
        d = f"-{d}"
    else:
        d = f"+{d}"

    m = f"{m:02d}"

    sex = [d, m, dec]
    return sex


def ra2sex(RA):
    """Convert RA decimal coordinate to sexagesimal.

    Parameters
    ----------
        RA : float The decimal angle

    Return
    ------
        str : The sexagesimal angle.

    """
    if math.isnan(RA):
        return None
    RA = RA / 15.0  # To hours
    sex = makeSex(RA)
    return f"{sex[0]}h {sex[1]}m {sex[2]:0.2f}s"


def dec2sex(dec):
    """Convert dec decimal coordinate to sexagesimal.

    Parameters
    ----------
        dec : float The decimal angle

    Return
    ------
        str : The sexagesimal angle.

    """
    if math.isnan(dec):
        return None
    sex = makeSex(dec)
    return f"{sex[0]}d {sex[1]}m {sex[2]:0.1f}s"


def plotLightCurve(
    lcData,
    xlog=False,
    ylog=False,
    whichCurves=None,
    fileName=None,
    T0=None,
    xlabel=None,
    ylabel=None,
    cols={},
    clobber=False,
    fig=None,
    ax=None,
    silent=True,
    verbose=False,
    oldAPI=False,
):
    """Create a plot of the light curve, if retrieved.

    This creates a basic plot of the suppied light curve, following
    normal XRT procedures.

    Parameters
    ----------

    lcData : dict
        A 'light curve ``dict``', in the standard format of the ukssdc
        module.

    xlog : bool, optional
        Whether to plot with the x axis logarithmic (default: ``False``)

    ylog : bool, optional
        Whether to plot with the y axis logarithmic (default: ``False``)

    whichCurves : list or tuple, optional
        If supplied, contains a subset of the light curve datasets to
        plot. NB, this will still interact with ``nosys`` and ``incbad``
        so if you set ``whichCurves=('PC_incbad',)`` but also
        ``incbad=False``, no data will be plotted.

    fileName : str, optional
        If ``None``, the light curve is plotted to screen, otherwise
        the light curve plot is saved to the specified file
        (default: ``None``).

    T0 : float, optional
        The T0 value, to use on the x-axis label.

    xlabel : str, optional
        The x-axis label (will be auto-set if not supplied.)

    ylabel : str, optional
        The y-axis label (will be auto-set if not supplied.)

    cols : dict, optional
        The colours to use for the data; should have keys: WT and PC.

    clobber : bool, optional
        Whether to overwrite the file if it exists (default ``False``).

    fig: matplotlib.figure.Figure, optional:
        A `matplotlib.figure.Figure` instance to use for the plotting.

    ax : matplotlib.axes._subplots.AxesSubplot, optonal
        A `matplotlib.axes._subplots.AxesSubplot` instance to use for
        plotting.

    silent : bool
        Whether to suppress all console output (default: ``True``).

    verbose : bool
        Whether to give verbose output for everything
        (default: ``False``).

    oldAPI : bool, optional
        Internal parameter, identifying if the lc columns are from an
        old version of the API.

    Returns
    -------
    fix,ax
        The objects containing the figure, created by
        ``matplolib.pyplot.subplots()``.

    """
    if verbose:
        silent = False

    if lcData is None:
        raise RuntimeError("No light curve has been supplied")

    if "Datasets" not in lcData:
        raise ValueError("Supplied lcData is not a light curve dict.")

    if whichCurves is None:
        whichCurves = lcData["Datasets"]

    if (fileName is not None) and (not clobber) and os.path.exists(fileName):
        raise RuntimeError(f"File `{fileName}` exists and clobber is False.")

    import matplotlib.pyplot as plt
    import numpy as np

    if (fig is None) or (ax is None):
        if verbose:
            print("Creating new subplot object")
        fig, ax = plt.subplots()

    # Another stupid hack because of bad old column names
    tneg = "TimeNeg"
    tpos = "TimePos"
    rneg = "RateNeg"
    rpos = "RatePos"
    ulCol = "UpperLimit"
    if oldAPI:
        tneg = "T_-ve"
        tpos = "T_+ve"
        rneg = "Rateneg"
        rpos = "Ratepos"
        ulCol = "Rate"

    hadRate = False
    hadHR = False

    if "WT" not in cols:
        cols["WT"] = "blue"

    if "PC" not in cols:
        cols["PC"] = "red"

    if "other" not in cols:
        cols["other"] = "red"

    # Have to work out which PC and WT datasets to use.
    for ds in whichCurves:
        if ds in cols:
            colour = cols[ds]
            label = ds
        elif "wt" in ds.lower():
            colour = cols["WT"]
            label = "WT"
        elif "pc" in ds.lower():
            colour = cols["PC"]
            label = "PC"
        else:
            colour = cols["other"]
            label = ds

        if "UL" in ds:
            hadRate = True
            if not silent:
                print(f"Plotting {ds} as upper limits")
            empty = np.zeros(len(lcData[ds]["Time"]))
            ulSize = np.zeros(len(lcData[ds]["Time"])) + 0.002
            ax.errorbar(
                lcData[ds]["Time"],
                lcData[ds][ulCol],
                xerr=[-lcData[ds][tneg], lcData[ds][tpos]],
                yerr=[ulSize, empty],
                uplims=True,
                elinewidth=1.0,
                color=colour,
                label=label,
                zorder=5,
                fmt="none",
            )

        elif "HR" in ds:
            hadHR = True
            if not silent:
                print(f"Plotting {ds} as a hardness ratio")
            trneg = "HRNeg"
            trpos = "HRPos"
            if (trneg not in lcData[ds].columns) and ("HRErr" in lcData[ds].columns):
                trneg = "HRErr"
                trpos = "HRErr"

            ax.errorbar(
                lcData[ds]["Time"],
                lcData[ds]["HR"],
                xerr=[-lcData[ds][tneg], lcData[ds][tpos]],
                yerr=[-lcData[ds][trneg], lcData[ds][trpos]],
                fmt=".",
                elinewidth=1.0,
                color=colour,
                label=label,
                zorder=5,
            )

        else:
            hadRate = True
            if not silent:
                print(f"Plotting {ds} as rates")
            trneg = rneg
            trpos = rpos
            if (trneg not in lcData[ds].columns) and ("RateErr" in lcData[ds].columns):
                trneg = "RateErr"
                trpos = "RateErr"

            ax.errorbar(
                lcData[ds]["Time"],
                lcData[ds]["Rate"],
                xerr=[-lcData[ds][tneg], lcData[ds][tpos]],
                yerr=[-lcData[ds][trneg], lcData[ds][trpos]],
                fmt=".",
                elinewidth=1.0,
                color=colour,
                label=label,
                zorder=5,
            )

    if xlog:
        ax.set_xscale("log")
    if ylog:
        ax.set_yscale("log")

    # See if we can make some nice axis labels
    if xlabel is None:
        xlabel = lcData["TimeFormat"]
        if "met" in lcData["TimeFormat"].lower():
            if T0 is not None:
                xlabel = f"Time since MET={T0} (s)"
            else:
                xlabel = "Time (s)"

    if ylabel is None:
        ylabel = "Count rate"
        if hadHR:
            if hadRate:
                ylabel = "Count rate or HR"
            else:
                ylabel = "Hardness ratio"

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    if fileName is None:
        plt.show()
    else:
        if verbose:
            print(f"Saving file {fileName}")
        plt.savefig(fileName)

    return fig, ax


def bayesRate(N, B, conf):
    """Calculate source brightness using Krafr, Burrows & Nousek 1991.

    This function receives the number of measured counts from some
    extraction, the number of expected background counts in the same
    location and a confidence level; it then return the mean number of
    source counts and the lower and upper bounds at the supplied
    confidence level.


    Parameters
    ----------

    N : int
        The number of counts in the source region.

    B : float
        The expected number of background counts.

    conf : float
        The confidence interval

    Returns
    -------
    list
        (min, max, mean)

    """
    if conf > 1:
        conf = conf / 100.0

    srcdim = 1000000
    srcstp = 0.001

    if N == 0:
        nlower = 0
        nupper = int(B * 1.1)
        if nupper < 30:
            nupper = 30
    else:
        tmp = int(N - B)
        if tmp < 0:
            tmp = 0
        nlower = tmp - 20 * tmp**0.5
        if nlower < 0:
            nlower = 0
        nupper = N + 20 * N**0.5
        if nupper < B:
            nupper = int(B * 1.1)
            if nupper < 30:
                nupper = 30

    mean = B + np.arange(nlower, nupper, srcstp)
    srcdim = len(mean) - 1

    # Calculate probability distribution
    if N > 0:
        alpha = -mean + N * np.log(mean) - math.log(math.factorial(N))
    else:
        alpha = -mean
    psrc = np.exp(alpha)

    # Peak and index of peak
    pmax = max(psrc)
    nmax = psrc.tolist().index(pmax)

    # Normalize distribution
    t = sum(psrc) - psrc[0]
    area = ((psrc[0] + psrc[srcdim]) / 2.0) + t
    t = 0.0
    psrc = psrc / area

    nupper = nmax
    nlower = nmax
    area = 0.0

    for i in [1, 2]:
        if psrc[nlower] == 0:
            nupper = nupper + 1
        elif psrc[nlower - 1] >= psrc[nupper + 1]:
            nlower = nlower - 1
        else:
            nupper = nupper + 1

    # By starting at the most probable
    # value and integrating in both directions,
    # always choosing to sum the side with the higher
    # probability, we can guarantee that the
    # confence interval chosen is the smallest
    # possible one.

    area = (psrc[nupper] + psrc[nlower]) / 2 + psrc[nlower + 1]

    for i in range(srcdim - 3):
        if nlower == 0:
            nupper = nupper + 1
            area = area + (psrc[nupper - 1] + psrc[nupper]) / 2
        else:
            if psrc[nlower - 1] >= psrc[nupper + 1]:
                nlower = nlower - 1
                area = area + (psrc[nlower + 1] + psrc[nlower]) / 2
            else:
                nupper = nupper + 1
                area = area + (psrc[nupper - 1] + psrc[nupper]) / 2

        if area >= conf:
            # Upper and lower limits are converted
            # into source rates.
            Smax = mean[nupper] - B  # nupper * srcstp
            Smin = mean[nlower] - B  # * srcstp
            return [Smin, Smax, nmax * srcstp]


def mergeLightCurveBins(
    lc,
    rows=None,
    remove=False,
    insert=False,
    forceRate=False,
    forceUL=False,
    ulConf=0.997,
    detThresh=None,
    silent=True,
    verbose=False,
):
    """Merge bins in a light curve.

    This function allows you to combine bins (rows) in a light curve;
    this is not the same as rebinning, where all bins are recalcuated,
    but rather for merging a subset of data. For example, if a
    light curve shows a series of consecutive upper limits that you want
    to merge, while leaving the rest of the light curve unaffected.

    This uses the Bayesian (Kraft, Burrows & Nousek 1991) method to
    find rates / limits, so if bins that combine to contain large
    numbers of counts are provided, performance may be poor. This may be
    reconsidered in future if use cases for merging bins with many
    counts are presented to the developers.

    The default behaviour of this function is to return a pandas Series
    containing the entry for the new bin, and to remove the old bins
    (that have been merged) from the light curve that you supplied.
    You can modify this using the arguments below, using ``remove`` to
    determine whether the merged bins are removed, and ``insert`` to
    add the new row to your light curve.

    NOTE: By default, this function will decide for itself whether to
    produce a (3-sigma) upper limit, or a count-rate with (1-sigma)
    error-bars, depending on whether then new bin has a 3-sigma lower
    limit above zero. You can determine what was produced by checking
    whether the returned Series contains the 'UpperLimit' or 'Rate' key.
    If you specify ``insert=True`` this will not happen, instead the bin
    will be forced to match the supplied light curve, i.e. it will be an
    upper limit if the light curve was one of upper limits, or a rate
    otherwise -- regardless of whether the bin constitutes a detetions.
    You can also use the ``force*`` arguments to control the type of bin
    created, but these are ignored if ``insert`` is ``True``.

    NOTE: The sigma value is not set by this method, as it requires
    knowledge of the relative sizes of the source and background regions
    which is not contained in the light curve file.

    Parameters
    ----------

    lc : pandas.DataFrame
        The light curve to work on

    rows : pandas.Series, optional
        A pandas series defining the rows to merge. If not supplied,
        all are merged.

    remove : bool, optional
        Whether to remove the merged lines from the light curve, as
        well as adding the new result (default: ``False``).

    insert : bool or str, optional
        Whether to add the new entry to the light curve. If ``True``
        the entry will be added to your light curve DataFrame.
        If ``False``, it will not be added, and this function will
        return the new entry.
        If 'match' it will only be added if it is the right `type`, i.e.
        if the merged bin is an upper limit, it will be added if the
        original light curve was of upper limits and the converse for a
        detection.

        **Warning** if ``insert`` is ``True`` then the type of bin
        created from the merge (rate or upper limit) will be forced to
        match that of the supplied light curve - and no test is done to
        determine if this is appropriate.

        (Default: ``False``).

    forceRate : bool, optional
        Return a count-rate and 1-sigma error, whether the data
        constitute a detection or not (default: ``False``).

    forceUL : bool, optional
        Return a 3-sigma upper limits, whether the data constitute a
        detection or not (default: ``False``).

    ulConf : float, optional
        The confidence level at which the upper limit should be
        determined, should be a probability (0-1); if >1 is assumed to
        be a percentage  (default: 0.997).

    detThresh : float or None, optional
        The probability threshold at which the count-rate is >0 for a
        detection to be determined (only used if ``detectionsAsRates``
        is ``True``). If ``None`` then this is set to ``conf``
        (default: ``None``).

    silent : bool
        Whether to suppress all console output (default: ``True``).

    verbose : bool
        Whether to give verbose output for everything
        (default: ``False``).


    Returns
    -------
    list
        The returned list has 3 entries (isUL, inserted, newData)

        'isUL' is a `bool` indicating whether the new bin is an upper
        limit.

        'inserted' is a `bool` indicating whether the new bin has been
        inserted into the supplied light curve.

        'newData' is a pandas Series containing the new bin.

    """
    if verbose:
        silent = False

    if forceRate and forceUL:
        raise RuntimeError("You can't force an upper limit and a rate!")

    if isinstance(insert, str):
        insert = insert.lower()
    if (insert is not True) and (insert is not False) and (insert != "match"):
        raise ValueError("Insert must be a boolean or 'match")

    if (insert is True) and (forceRate or forceUL) and not silent:
        if forceRate:
            print("WARNING: Ignoring `forceRate` as insert=True")
        else:
            print("WARNING: Ignoring `forceUL` as insert=True")

    if not isinstance(ulConf, float):
        raise ValueError("`conf` parameter must be a float!")
    if ulConf > 1:
        if verbose:
            print(f"Interpreting conf={ulConf} as a percentage")
        ulConf = ulConf / 100.0

    dtIsC = False

    if detThresh is None:
        detThresh = ulConf
        dtIsC = True
    else:
        if not isinstance(detThresh, float):
            raise ValueError("`detThresh` parameter must be a float!")
        if detThresh > 1:
            if verbose:
                print(f"Interpreting detThresh={detThresh} as a percentage")
            detThresh = detThresh / 100.0

    # If we have upper limits we have to weight CF by exposure,
    # otherwise by counts.
    isOrigUL = False
    sumCol = "CtsInSrc"
    if "UpperLimit" in lc.columns:
        isOrigUL = True
        sumCol = "Exposure"

    tmp = []
    if rows is None:
        tmp = lc.copy()
    else:
        tmp = lc[rows].copy()

    # Need to calculate some new columns.
    # CFE = CF * E, will need this for the new CF
    tmp["CFE"] = tmp["CorrFact"] * tmp[sumCol]
    # Start and stop to get new time
    tmp["START"] = tmp["Time"] + tmp["TimeNeg"]
    tmp["STOP"] = tmp["Time"] + tmp["TimePos"]
    # Need BGRate and error.
    # These will be the current bins' exposure-weighted mean, but for
    # error should add in quadature, so need err**2
    tmp["BGRSQ"] = tmp["BGerr"] ** 2
    tmp["BGE"] = tmp["BGrate"] * tmp["Exposure"]
    tmp["BGErrE"] = tmp["BGRSQ"] * tmp["Exposure"]

    # Now calculate the values for the new bin
    E = tmp["Exposure"].sum()
    C = int(tmp["CtsInSrc"].sum())
    B = tmp["BGInSrc"].sum()
    CFE = tmp["CFE"].sum()
    weight = tmp[sumCol].sum()
    CF = CFE / weight

    # And time properities

    meanT = tmp["Time"].mean()
    startT = tmp["START"].min()
    stopT = tmp["STOP"].max()
    dur = stopT - startT
    tPos = stopT - meanT
    tNeg = startT - meanT
    fracExp = E / dur

    # And background properties
    bgRate = tmp["BGE"].sum() / E
    bgErr = math.sqrt(tmp["BGErrE"].sum()) / E

    # # And remove the temporary columns
    # NO NEED - they are in tmp. Only reinstate this if I need to.
    # lc.drop(["CFE"], axis=1, inplace=True)
    # lc.drop(["START"], axis=1, inplace=True)
    # lc.drop(["STOP"], axis=1, inplace=True)

    # Do we need a rate or UL? Or is it up to me to decide?
    # If force* is set, we do not decide.
    # If insert it set, we do not decide.
    # Otherwise we do.
    # Let's some some variables also to track what we need, and what
    # we got
    getUL = None
    UL = None
    gotUL = False
    if insert is True:
        getUL = isOrigUL
    elif forceRate:
        getUL = False
    elif forceUL:
        getUL = True
    else:
        if verbose:
            print("Checking whether the new bin is a detection.")
        (smin, smax, mean) = bayesRate(C, B, detThresh)
        if dtIsC:
            gotUL = True
        UL = smax * CF / E
        if smin > 0:
            getUL = False
            if not silent:
                print("The new bin is a detection")
        else:
            getUL = True
            if not silent:
                print("The new bin is an upper limit")

    # Create variable for the new bin
    newRow = {
        "Time": meanT,
        "TimePos": tPos,
        "TimeNeg": tNeg,
        "RatePos": 0,
        "RateNeg": 0,
        "FracExp": fracExp,
        "BGRate": bgRate,
        "BGErr": bgErr,
        "CorrFact": CF,
        "CtsInSrc": C,
        "BGInSrc": B,
        "Exposure": E,
        "Sigma": math.nan,
        "SNR": 0,
    }

    newIndex = [
        "Time",
        "TimePos",
        "TimeNeg",
        "HOLDER",
        "RatePos",
        "RateNeg",
        "FracExp",
        "BGRate",
        "BGErr",
        "CorrFact",
        "CtsInSrc",
        "BGInSrc",
        "Exposure",
        "Sigma",
        "SNR",
    ]

    # Are we finding an UL?
    if getUL:
        # If we haven't already found it, get it.
        if not gotUL:
            (smin, smax, mean) = bayesRate(C, B, ulConf)
            UL = smax * CF / E
        # Create new row with UL
        newRow["UpperLimit"] = UL
        newIndex[3] = "UpperLimit"

    # If we need a rate, get it:
    if not getUL:
        (smin, smax, mean) = bayesRate(C, B, 0.683)
        rate = mean * CF / E
        neg = (smin - mean) * CF / E
        pos = (smax - mean) * CF / E
        snr = rate / abs(neg)
        # Create new row with rate
        newRow["Rate"] = rate
        newRow["RatePos"] = pos
        newRow["RateNeg"] = neg
        newRow["SNR"] = snr
        newIndex[3] = "Rate"

    newData = pd.Series(newRow, index=newIndex)

    if remove:
        lc.drop(tmp.index, inplace=True)

    # If insert is True we save this row.
    # If insert is "match" we save it if it is the same format as the
    # existing table.
    inserted = False
    if (insert is True) or ((insert == "match") and (getUL == isOrigUL)):
        if verbose:
            print("Inserting merged bin into the light curve.")
        lc.loc[-1] = newData
        inserted = True

    lc.sort_values(
        by=[
            "Time",
        ],
        axis=0,
        inplace=True,
    )

    lc.reset_index(drop=True, inplace=True)

    return (getUL, inserted, newData)


def mergeUpperLimits(
    ultab, rows=None, detectionsAsRates=True, bands="all", conf=0.997, detThresh=None, silent=True, verbose=False
):
    """Merge rows in an upper limit table returned.

    This function allows you to combine rows in a set of upper limits
    returned by the uds.getUpperLimits() function.

    This uses the Bayesian (Kraft, Burrows & Nousek 1991) method to
    find rates / limits, so if bins that combine to contain large
    numbers of counts are provided, performance may be poor. This may be
    reconsidered in future if use cases for merging bins with many
    counts are presented to the developers.

    This returns a single `dict` with the upper limit and/or rate,
    along with the error and the details of the counts, bg counts etc.

    Parameters
    ----------

    ultab : pandas.DataFrame
        The upper limit result to work on

    rows : pandas.Series, optional
        A pandas series defining the rows to merge. If not supplied,
        all are merged.

    detectionsAsRates : bool, optional
        Whether to check if the source is actually detected (i.e. has a
        lower confidence bound >0) in the merged result. If ``True``
        then, the {band}Rate, {band}_RatePos, {band}_RateNeg and
        {band}_RateIsDetected columns will be returned; the former three
        will be NaN  if the source is not detected (default: `True``).

    bands : str or list or tuple, optional
        Which bands to calculate the merged result for. If 'all' then
        all bands are processed. Note: if bands are selected but not
        supplied in the ultab, they will be skipped.

    conf : float, optional
        The confidence level at which the upper limit should be
        determined, should be a probability (0-1); if >1 is assumed to
        be a percentage  (default: 0.997).

    detThresh : float or None, optional
        The probability threshold at which the count-rate is >0 for a
        detection to be determined (only used if ``detectionsAsRates``
        is ``True``). If ``None`` then this is set to ``conf``
        (default: ``None``).

    silent : bool
        Whether to suppress all console output (default: ``True``).

    verbose : bool
        Whether to give verbose output for everything
        (default: ``False``).


    Returns
    -------
    dict
        The dictionary of results.

    """
    if verbose:
        silent = False

    if not isinstance(conf, float):
        raise ValueError("`conf` parameter must be a float!")
    if conf > 1:
        if verbose:
            print(f"Interpreting conf={conf} as a percentage")
        conf = conf / 100.0

    if detThresh is None:
        detThresh = conf
    else:
        if not isinstance(detThresh, float):
            raise ValueError("`detThresh` parameter must be a float!")
        if detThresh > 1:
            if verbose:
                print(f"Interpreting detThresh={detThresh} as a percentage")
            detThresh = detThresh / 100.0

    useBands = []
    if isinstance(bands, str):
        if bands.lower() == "all":
            useBands = SXPS_BAND_NAMES
        else:
            tb = bands[0].upper() + bands[1:].lower()
            if tb not in SXPS_BAND_NAMES:
                raise ValueError(f"Band `{tb}` is not recognised")
            useBands = [
                tb,
            ]
    elif isinstance(bands, (tuple, list)):
        for b in bands:
            tb = b[0].upper() + b[1:].lower()
            if tb not in SXPS_BAND_NAMES:
                raise ValueError(f"Band `{tb}` is not recognised")
            useBands.append(tb)

    tmp = []
    if rows is None:
        tmp = ultab.copy()
    else:
        tmp = ultab[rows].copy()

    needCols = ("SourceExposure", "ImageExposure")
    for c in needCols:
        if c not in tmp.columns:
            raise ValueError(f"Column `{c}` is mandatory but not in your supplied data frame")
    totSourceExp = tmp["SourceExposure"].sum()
    totImExp = tmp["ImageExposure"].sum()

    ret = {"SourceExposure": totSourceExp, "ImageExposure": totImExp}

    needCols = ("Counts", "BGCounts", "CorrectionFactor")
    for b in useBands:
        # First, do we have all of the columns?
        gotAll = True
        for c in needCols:
            if f"{b}_{c}" not in tmp.columns:
                gotAll = False
                break
        if not gotAll:
            if not silent:
                print(f"Skipping band `{b}` as data are missing")
            continue

        tmp[f"{b}_CFE"] = tmp[f"{b}_CorrectionFactor"] * tmp["ImageExposure"]

        totC = int(tmp[f"{b}_Counts"].sum())
        totB = tmp[f"{b}_BGCounts"].sum()
        CF = tmp[f"{b}_CFE"].sum() / totImExp

        # Don't want to do unneeded calls.
        # So, if we have to check det, do that first.
        # If it is undetected, and conf == detThresh then no need to
        # recalc.
        # Otherwise, getUL
        # Get Rate if rates are req or detected.
        ul = None
        isDet = False
        rate = math.nan
        ratePos = math.nan
        rateNeg = math.nan
        if detectionsAsRates:
            # Check detection
            (smin, smax, mean) = bayesRate(totC, totB, detThresh)
            # If detThresh == conf then we've also already got the UL so save it
            if detThresh == conf:
                ul = smax * CF / totImExp

            # Now get the rate if we need it
            if smin > 0:  # Detection
                if not silent:
                    print("Merge gives a detection.")
                isDet = True
                # Get rate and errors:
                (smin, smax, mean) = bayesRate(totC, totB, 0.683)
                rate = mean * CF / totImExp
                rateNeg = (smin - mean) * CF / totImExp
                ratePos = (smax - mean) * CF / totImExp
            elif not silent:
                print(f"Merge does not give a detection in band {b}.")

        if ul is None:  # Haven't got UL. Always get UL
            (smin, smax, mean) = bayesRate(totC, totB, conf)
            ul = smax * CF / totImExp
        ret[f"{b}_UpperLimit"] = ul
        ret[f"{b}_Counts"] = totC
        ret[f"{b}_BGCounts"] = totB
        ret[f"{b}_CorrectionFactor"] = CF
        if detectionsAsRates:
            ret[f"{b}_Rate"] = rate
            ret[f"{b}_RatePos"] = ratePos
            ret[f"{b}_RateNeg"] = rateNeg
            ret[f"{b}_IsDetected"] = isDet

    return ret



__docformat__ = "restructedtext en"

from .. import main as base
from ..data import download as dl
import pandas as pd
from .DBFilter import filter

if base.HAS_ASTROPY:
    import astropy.coordinates
    import astropy.units
if base.HAS_ASTROQ:
    import astroquery.simbad as aqs
    import astroquery.vizier as aqv

MAXROWS = 1000


class dataQuery:
    """The base case for UKSSDC data requests. A 'virtual' class.

    This class should never in itself be instantiated, only those
    dervied from it. In fact, it will contain enough functionality to
    run for the most basic of cases, except that it will never select
    which table / catalogue is being searched.

    """

    def __init__(self, silent=True, verbose=False):
        """Create a dataQuery instance.

        Parameters
        ----------
        silent : bool
            Whether to suppress all console output (default: ``True``).

        verbose : bool
            Whether to give verbose output for everything
            (default: `False`; overridden by silent).

        """
        # This 'abstract' class has no table defined.
        self._dbName = None
        self._table = None
        self._silent = silent
        self._verbose = verbose
        self._metadata = None
        self._colsToGet = None
        self._defaultCols = None
        self._defaultColSets = None
        self._tables = []
        self._coneRA = None
        self._coneDec = None
        self._coneName = None
        self._coneRadius = None
        self._coneUnits = "arcsec"
        self._doConeSearch = False
        self._filters = []
        self._sortCol = None
        self._sortDir = "ASC"
        self._results = None
        # numRows and firstRows, if None, means no limits
        self._maxRows = MAXROWS
        self._firstRow = None
        self._numRows = None
        self._resolvedInfo = None
        self._resolvedRA = None
        self._resolvedDec = None
        self._locked = False
        self._raw = None  # DEBUG PROPERTY, DELETE LATER
        self._obsCol = None
        self._targetCol = None
        self._obsIDAsString = False
        self._defaultIDCol = None
        self._raCol = None
        self._decCol = None

        self._prodData = {"SIMBAD": None, "Vizier": None}

        if self._verbose:
            self._silent = False
            print("Disabling silent mode, verbose mode was requested.")

    # End of __init__

    # -----------------------------------------------------------------
    # Now set up variable access, via properties to control read/write
    # etc.

    # Silent
    @property
    def silent(self):
        """Whether to suppress output."""
        return self._silent

    @silent.setter
    def silent(self, silent):
        if not isinstance(silent, bool):
            raise ValueError("Silent must be a bool")
        self._silent = silent

    # Verbose
    @property
    def verbose(self):
        """Whether to write extra output."""
        return self._verbose

    @verbose.setter
    def verbose(self, verbose):
        if not isinstance(verbose, bool):
            raise ValueError("Verbose must be a bool")
        self._verbose = verbose
        if verbose:
            self.silent = False

    # make setting table unimplemented here, allowing the subclasses to
    # decide whether you can change table mid query; also some classes
    # may only support one table so they would want these functions not
    # to work (i.e. set self._table in constructor.
    # table
    @property
    def table(self):
        """Which database table to query."""
        return self._table

    @table.setter
    def table(self, table):
        self.checkLock()
        if table not in self._tables:
            raise ValueError(f"{table} is not a valid table. The list `tables` shows valid values.")
        self._table = table
        self._metadata = None
        # if (self._defaultColSets is not None) and table in (self._defaultColSets):
        #     self._defaultCols = self._defaultColSets[table]
        # If unlocking the table, we have to forget results as various results handling functions are tied to the table
        # that was used.
        self._results = None
        self.reset()
        if self.verbose:
            print(f"Selecting table `{table}`")

    # dbName is also unset here, may not be changeable depending on sub-class.
    # dbName
    @property
    def dbName(self):
        """Which database is to be queried."""
        if self._dbName is None:
            raise NotImplementedError
        return self._dbName

    @dbName.setter
    def dbName(self, dbName):
        raise NotImplementedError

    # coneRA
    @property
    def coneRA(self):
        """The RA on which to centre a cone search (J2000)."""
        return self._coneRA

    @coneRA.setter
    def coneRA(self, RA):
        self.checkLock()
        gotRA = False
        if base.HAS_ASTROPY:
            if isinstance(RA, astropy.coordinates.Angle):
                RA = float(RA.deg)
                gotRA = True
            elif isinstance(RA, str):
                tmp = astropy.coordinates.Angle(RA)
                RA = float(tmp.deg)
                gotRA = True
        # If we don't have astropy, or it wasn't something astropy has
        # parsed, then it must be int or float and we will parse it
        if not gotRA:
            if not isinstance(RA, (int, float)):
                raise ValueError("RA must be int or float")
        self._coneRA = float(RA)

    # coneDEC
    @property
    def coneDec(self):
        """The Dec on which to centre a cone search (J2000)."""
        return self._coneDec

    @coneDec.setter
    def coneDec(self, Dec):
        self.checkLock()
        gotDec = False
        if base.HAS_ASTROPY:
            if isinstance(Dec, astropy.coordinates.Angle):
                Dec = float(Dec.deg)
                gotDec = True
            elif isinstance(Dec, str):
                tmp = astropy.coordinates.Angle(Dec)
                Dec = float(tmp.deg)
                gotDec = True
        # If we don't have astropy, or it wasn't something astropy has
        # parsed, then it must be int or float and we will parse it
        if not gotDec:
            if not isinstance(Dec, (int, float)):
                raise ValueError("Dec must be int or float")
        self._coneDec = float(Dec)

    # coneName
    @property
    def coneName(self):
        """The name of the object on which to centre a cone search."""
        return self._coneName

    @coneName.setter
    def coneName(self, name):
        self.checkLock()
        if not isinstance(name, str):
            raise ValueError("Name must be a string")
        self._coneName = name

    # coneRadius
    @property
    def coneRadius(self):
        """The Radius on which to centre a cone search."""
        return self._coneRadius

    @coneRadius.setter
    def coneRadius(self, Radius, Units=None):
        self.checkLock()
        if not isinstance(Radius, (int, float)):
            raise ValueError("Radius must be int or float")
        self._coneRadius = float(Radius)
        if Units is not None:
            self.coneUnits = Units

    # coneUnits
    @property
    def coneUnits(self):
        """The units of the cone-search radius."""
        return self._coneUnits

    @coneUnits.setter
    def coneUnits(self, Units):
        self.checkLock()
        if not isinstance(Units, str):
            raise ValueError("Units must be a string.")
        self._coneUnits = Units

    # maxRows
    @property
    def maxRows(self):
        """How many rows to retrieve. None=all."""
        self.checkLock()
        return self._maxRows

    @maxRows.setter
    def maxRows(self, num):
        if not isinstance(num, (int, float)) and (num is not None):
            raise ValueError("num must be a number or None")
        self._maxRows = num

    # firstRow
    @property
    def firstRow(self):
        """The first row to retrieve. None=auto."""
        return self._firstRow

    @firstRow.setter
    def firstRow(self, num):
        self.checkLock()
        if (not isinstance(num, (int, float))) and (num is not None):
            raise ValueError("num must be a number or None")
        self._firstRow = num

    @property
    def ObsIDAsString(self):
        """Whether ObsIDs should be converted to strings."""
        return self._obsIDAsString

    @ObsIDAsString.setter
    def ObsIDAsString(self, val):
        if not isinstance(val, bool):
            raise ValueError("val must be a boolean.")
        self._obsIDAsString = val

    # sortCol
    @property
    def sortCol(self):
        """Which column to sort on"""
        return self._sortCol

    @sortCol.setter
    def sortCol(self, colName):
        if not isinstance(colName, str):
            raise ValueError("colName must be a string.")
        self.checkLock()
        if colName not in self.metadata["ColName"].values:
            raise ValueError(f"`{colName}` is not a valid column name.")
        self._sortCol = colName

    # sortDir
    @property
    def sortDir(self):
        """Which direction to sort in"""
        return self._sorDir

    @sortDir.setter
    def sortDir(self, dir):
        if not isinstance(dir, str):
            raise ValueError("dir must be a string.")
        self.checkLock()
        dir = dir.upper()
        if dir not in ("ASC", "DESC"):
            raise ValueError("Sort dir must be 'ASC' or 'DESC'")
        self._sortDir = dir

    # ---------------------------
    # And some which we want to be read-only when accessed as a variable.

    # doConeSearch
    @property
    def doConeSearch(self):
        """Whether a cone search will be done."""
        return self._doConeSearch

    # metadata
    @property
    def metadata(self):
        """The metadata for with the selected table. Read-only."""
        # Need to get metadata if we don't have it:
        if self._metadata is None:
            if not self.silent:
                print("Need to get the metadata.")
            self.getMetadata()

        return self._metadata

    # colsToGet
    @property
    def colsToGet(self):
        """The columns selected for retrieval."""
        return self._colsToGet

    # filters
    @property
    def filters(self):
        """The filters that will be applied."""
        return self._filters

    # Results
    @property
    def results(self):
        """The results of the query."""
        return self._results

    # haveResults
    @property
    def haveResults(self):
        """Whether we have results from this query."""
        return self._results is not None

    @property
    def locked(self):
        """Is this object locked?"""
        return self._locked

    @property
    def numRows(self):
        """The number of rows returned by the query."""
        return self._numRows

    @property
    def resolvedInfo(self):
        """The ouput of the name resolver."""
        return self._resolvedInfo

    @property
    def resolvedRA(self):
        """The RA returned by the name resolver."""
        return self._resolvedRA

    @property
    def resolvedDec(self):
        """The dec returned by the name resolver."""
        return self._resolvedDec

    @property
    def tables(self):
        """The tables that can be selected."""
        return self._tables

    @property
    def defaultCols(self):
        """The columns that are retrieved if no selection is created."""

        if self._defaultCols is None:
            # self._checkMetaData()
            if "Class" in self.metadata.columns:
                # Can do this all in one line, but it's a bit hard to read, so lets be nice:
                # First, filter the metadata on cases where Class has "BASIC" in it
                tmp = self.metadata.loc[self.metadata["Class"].str.contains("BASIC")]
                # Now get the columns in this:
                self._defaultCols = tmp["ColName"].tolist()

        return self._defaultCols

    @property
    def cols(self):
        """Columns in this table."""
        return self.metadata["ColName"].values

    @property
    def obsColumn(self):
        """Which column contains the observation identifier."""
        return self._obsCol

    @property
    def targetColumn(self):
        """Which column contains the target identifier."""
        return self._targetCol

    @property
    def raCol(self):
        """Which column is used as RA for cone searches."""
        return self._raCol

    @property
    def decCol(self):
        """Which column is used as dec for cone searches."""
        return self._decCol

    @property
    def SIMBAD(self):
        """The SIMBAD search results."""
        return self._prodData["SIMBAD"]

    @property
    def Vizier(self):
        """The Vizier search results."""
        return self._prodData["Vizier"]

    # ---------------------------------------------------------------
    # Functions. First some standard things:

    def __str__(self):
        str = f"PRINTING AN {type(self)} object."
        return str

    def __repr__(self):
        str = f"I AM AN {type(self)} object."
        return str

    def checkLock(self):
        """Check if this object is locked."""
        if self._locked:
            raise RuntimeError("Cannot edit this request as it is locked.")

    def unlock(self):
        """Unlock the object for editing."""
        self._locked = False

    # ---------------------------------------------------------------
    # Metadata
    def getMetadata(self):
        """Retrieve the metadata for this catalogue from the server.

        This queries the server for the metadata associated with the
        database/table of the current object, and saves it into the
        metadata variable as a pandas object.

        """
        self.checkLock()
        sendData = {"database": self.dbName, "table": self.table}
        if self.verbose:
            print(f"Getting metadata for {self.dbName}.{self.table}")

        ret = base.submitAPICall("getMetadata", sendData, minKeys=["metadata"], verbose=self._verbose)

        # metadata should have two entries: 'columns' and 'data'
        self._metadata = pd.DataFrame(ret["metadata"]["metadata"], columns=ret["metadata"]["columns"])

        self._obsCol = None
        self._targetCol = None
        if "IsObsCol" in self._metadata:
            self._metadata["IsObsCol"] = pd.to_numeric(self._metadata["IsObsCol"])
            tmp = self._metadata.loc[self._metadata["IsObsCol"] == 1]["ColName"]
            if len(tmp) > 0:
                self._obsCol = tmp.iloc[0]
                if len(tmp) > 1 and not self.silent:
                    print(
                        "WARNING: Metadata contains TWO obs columns! This may be a bug; "
                        "please notify swift-help@leicester.ac.uk"
                    )

        if "IsTargetCol" in self._metadata:
            self._metadata["IsTargetCol"] = pd.to_numeric(self._metadata["IsTargetCol"])
            tmp = self._metadata.loc[self._metadata["IsTargetCol"] == 1]["ColName"]
            if len(tmp) > 0:
                self._targetCol = tmp.iloc[0]
                if len(tmp) > 1 and not self.silent:
                    print(
                        "WARNING: Metadata contains TWO target columns! This may be a bug; "
                        "please notify swift-help@leicester.ac.uk"
                    )

    # --------------------------------------------------------------
    # Misc
    def reset(self, keepCone=False, keepFilters=False, keepCols=False):
        """Reset this query.

        This resets the query so it can be redefined or rerun. By
        default, everything is reset, but you can override this if you
        wish.

        Parameters
        ----------

        keepCone : bool, optional
            Whether to keep the cone search filter (default: ``False``).

        keepFiters : bool, optional
            Whether to keep the defined filters (default: ``False``).

        keepCols : bool, optional
            Whether to keep the defined columns to retrieve
            (default: ``False``).

        """
        if not self.silent:
            print("Resetting query details")
        self.unlock()
        if not keepFilters:
            self.removeAllFilters()
        if not keepCone:
            self.removeConeSearch()
        if not keepCols:
            self._colsToGet = None
        self._resolvedInfo = None
        self._resolvedRA = None
        self._resolvedDec = None
        self._results = None
        self._sortCol = None
        self._metadata = None
        self._defaultCols = None

        self.clearProducts()
        self._raw = None  # TEMP LINE

    def clearProducts(self):
        """Clear all stored product data"""
        for k in self._prodData.keys():
            self._prodData[k] = None
            if self.verbose:
                print(f"Clearing `{k}`")

    def clearProduct(self, prod):
        """Clear specific product data.

        This clears the data stored internally corresponding to a
        specific product type (e.g. lightCurve).

        Parameters
        ----------

        prod : str
            The name of the product to clear.

        """
        if prod not in self._prodData:
            raise ValueError(f"Product `{prod}` is not recognised.")
        self._prodData[prod] = None
        if not self.silent:
            print(f"Cleared stored `{prod}` data.")

    # ---------------------------------------------------------------
    # Cone search functions
    # These allow a cone search to be build, or requested.
    # Note that simply setting coneRA/Dec above on their own doesn't force
    # a cone search to run unless doConeSearch is set. These methods are
    # preferred ways of managing the cone search.

    def addConeSearch(self, ra=None, dec=None, radius=None, name=None, position=None, units="arcsec"):
        """Include a cone search filter on the dataQuery.

        This function adds a cone search to the set of filters to be
        applied to your request. If a cone serach already existed, this
        will overwrite it; you cannot (at present) have multiple cone
        searches combined with an OR function; you will have to do
        multiple cone searches instead.

        Note: name, or ra & dec must be specified, but they cannot
        both be specified. If you supply both, an error will be raised.

        Parameters
        ----------
        ra : float or astropy.coordinates.Angle
            Central RA of the cone, in J2000.

        dec : float or astropy.coordinates.Angle
            Central Dec of the cone, in J2000.

        name : str
            The name of an object to centre on. Will be resolved.

        position : str
            Free-form coordinates to attempt to parse

        radius : float
            Radius of the cone search.

        units : str
            The units of coneRadius. Default 'arcsec'.

        """
        self.checkLock()
        # Set mainly via @property functions so that checks on values are done.
        if name is not None:
            if (ra is not None) or (dec is not None) or (position is not None):
                raise ValueError("You must supply name OR position, not both.")
            self._coneRA = None
            self._coneDec = None
            self.coneName = name
        elif position is not None:
            self.coneName = position
        elif (ra is None) or (dec is None):
            raise ValueError("You must supply name or position.")
        else:
            self.coneRA = ra
            self.coneDec = dec
            self._coneName = None

        if radius is None:
            raise ValueError("radius must be supplied")

        self.coneRadius = radius
        self.coneUnits = units

        self._doConeSearch = True

    def editConeSearch(self, coneRA, coneDec, coneRadius, coneUnits="arcsec"):
        """Changes the cone search. Just a wrapper to addConeSearch."""
        self.addConeSearch(coneRA, coneDec, coneRadius, coneUnits)

    def removeConeSearch(self):
        """Remove the cone search from the filters to apply."""
        self.checkLock()
        self._coneRA = None
        self._coneDec = None
        self._coneRadius = None
        self._coneUnits = "arcsec"
        self._doConeSearch = False

    def showConeSearch(self):
        """Print details of the current cone search."""

        if self._doConeSearch:
            if self._coneName is None:
                what = f"({self.coneRA}, {self.coneDec})"
            else:
                what = self._coneName
            print(f"{self.coneRadius} {self.coneUnits} around {what}")
        else:
            print("None")

    def isValid(self):
        """Whether the current request is valid and can be submitted."""

        # Need to get metadata if we don't have it:
        if self._metadata is None:
            if not self.silent:
                print("Need to get the metadata to check the query is valid.")
            self.getMetadata()

        # This needs to check:
        # All columns are permissable
        # If not set, check defaults as user could have been naughty.
        if self.verbose:
            print("Checking requested columns...")
        tmp = self._colsToGet
        if tmp is None:
            tmp = self.defaultCols
            if tmp is None:
                print("No columns selected to retrieve!")
                return False
        if tmp != "*":
            for c in tmp:
                if c not in self.metadata["ColName"].values:
                    if not self.silent:
                        print(f"Requested column {c} does not exist.")
                    return False

        # Now check filters
        # Filters
        if self.verbose:
            print("Checking filters...")
        for f in self._filters:
            if not f.isValid(self.metadata):
                return False

        # And check the cone search
        if self._doConeSearch:
            if self.verbose:
                print("Checking cone search parameters...")
            # Need a radius
            if self.coneRadius is None:
                if not self.silent:
                    print("A cone search is selected, but radius is not set")
                return False
            # Need name or ra AND dec
            if (self.coneName is None) and ((self.coneRA is None) or (self.coneDec is None)):
                if not self.silent:
                    print("A cone search is selected, but neither name, nor RA & Dec are set.")
                return False
            if (self.coneName is not None) and ((self.coneRA is not None) or (self.coneDec is not None)):
                if not self.silent:
                    print("You should supply name OR position for a cone search.")
                return False

        return True

    # ---------------------------------------------------------------
    # Column functions

    def _addCol(self, colName):
        """Internal function to add a column to the list.

        This is called by addCol() (which can support strings or lists)
        but must recieve a string. It adds the item to _colsToGet, after
        verifying it.

        The table metadata must be known, so it will retrieve it if it
        doesn't exist.

        Will raise ValueError if the supplied parameter is not a string,
        or is not a valid column (or '*').

        Parameters
        ----------

        colName : str
            The column to add.

        """
        self.checkLock()
        # First, check whether the metadata is retrieved and up to date.
        if not isinstance(colName, str):
            raise ValueError("colName should be a string")

        if colName == "*":
            if self.verbose:
                print("Setting to retrieve all columns.")
            self._colsToGet = self.metadata["ColName"].values.tolist()
        else:
            # Is the column name valid?
            if colName not in self.metadata["ColName"].values:
                raise ValueError(f"`{colName}` is not a valid column name.")
            # If previously we had selected all, then warn if not silent.
            if self._colsToGet == "*":
                if not self.silent:
                    print("WARNING: previously you were selecting all data; you are now requesting specific columns.")
                self._colsToGet = [
                    colName,
                ]

            # If this is the first column, create the list
            if self._colsToGet is None:
                self._colsToGet = [
                    colName,
                ]
                if self.verbose:
                    print(f"Will retrieve column {colName}")
            else:
                # This 'else' assumes a list, i.e. colsToGet is either '*', None or a list.
                # If it is not, then likely this will throw an error. That's OK, since if it is
                # not one of these then a user has edited this "hidden" field directly, and they
                # deserve what they get.
                if colName in self._colsToGet:
                    if not self.silent:
                        print(f"Cannot add column {colName}; it is already selected.")
                else:
                    self._colsToGet.append(colName)
                    if self.verbose:
                        print(f"Will retrieve column {colName}")

    def addCol(self, colName):
        """Add a column/columns to the list of those to retrieve.

        This can receive either a string, which is a column name or '*',
        or a list of column names to add to the list to retrieve.
        Note: '*' cannot appear in a list.

        The names are checked against valid column names, so if it has
        not already been called, getMetadata() will run first.

        Parameters
        ----------

        colName : str or list
            The column(s) to add.

        """
        self.checkLock()
        if isinstance(colName, str):
            self._addCol(colName)
        elif isinstance(colName, (list, tuple)):
            if "*" in colName:
                raise ValueError("You cannot include '*' in a list of columns.")
            for c in colName:
                self._addCol(c)
        else:
            raise ValueError("colName must be a string or list")

        if self.verbose:
            print(f"Set to retrieve columns: {self._colsToGet}")

    def removeAllCols(self):
        """Empty the list of defined columns to retrieve."""
        self.checkLock()
        self._colsToGet = None

    def _removeCol(self, colName):
        """Internal function to remove a column from the list.

        This is called by removeCol() (which supports strings or lists)
        but must recieve a string. It removes the item to _colsToGet.

        Parameters
        ----------

        colName : str
            The column to remove.

        """
        self.checkLock()
        if not isinstance(colName, str):
            raise ValueError("colName should be a string")

        self.colsToGet.remove(colName)

    def removeCol(self, colName):
        """Remove a column/columns to the list of those to retrieve.

        This can receive either a string, which is a column name or a
        list of column names to add to the list to retrieve.

        Parameters
        ----------

        colName : str or list
            The column(s) to add.

        """
        self.checkLock()
        if isinstance(colName, str):
            self._removeCol(colName)
        elif isinstance(colName, (list, tuple)):
            for c in colName:
                self._removeCol(c)
        else:
            raise ValueError("colName must be a string or list")

        if len(self._colsToGet) == 0:
            self._colsToGet = None

        if self.verbose:
            print(f"Will retrieve columns: {self._colsToGet}")

    # ----------------------
    # Filter functions

    # addFilter
    # removeFilter
    # showFilters (maybe with option for 'as SQL' vs as Dict)

    def removeAllFilters(self):
        """Remove all defined search filters."""
        self.checkLock()
        self._filters = []

    def addFilter(self, filterDef):
        """Add a filter to the query.

        This can be done two ways:

        1. By passing a dict with the following keys --
            some are optional:

            * colName - The name of the column to filter on
            * filter - The filter to apply ('<', '>', 'IS NULL' etc)
            * val - The value to apply to the filter, if appropriate.
                e.g. if the filter is '<', val may be 3.1234.
                If the filter takes no arguments (IS [NOT] NULL) this is
                ignored.
                If the filter takes two arguments (BETWEEN) this should
                be a list.
            * combiner - OPTIONAL --  This can be AND or OR, if this
                filter has two components.
            * filter2 - As filter, for the second clause.
            * val2 - As val, for the second clause.

        2. By passing a tuple, whose values are the above keys, in order,
            e.g. ('foo', '<', 3, 'OR', 'foo', 'BETWEEN', [7,10] )


        Parameters
        ----------

        filterDef : string or list or tuple
            The filter definition

        """
        self.checkLock()
        # colname is
        self._filters.append(filter(filterDef, self.metadata))

    def showFilters(self):
        """List all filters currently applied"""
        i = 0
        for f in self._filters:
            print(f"{i}:\t{f}")
            i = i + 1

    def removeFilter(self, ix):
        """Remove a filter, by index.

        To see filters and their indices, call showFilters

        Parameters
        ----------

        ix : int
            Index of filter to remove

        """
        self.checkLock()
        if not isinstance(ix, int):
            raise ValueError("ix must be an int")
        if (ix < 0) or (ix >= len(self._filters)):
            raise ValueError(f"ix must be between 0 and {len(self._filters) - 1}")
        del self._filters[ix]
        if not self._silent:
            self.showFilters()

    # ---------------------------------------------------------------
    # Actual search functions

    def submit(self, **kwargs):
        """Submit the query."""
        self.checkLock()

        # First, check validity. Do this by function call
        if not self.isValid():
            if not self.silent:
                print("Cannot submit query - it is not valid.")
            return False

        # Build the API request dict:
        sendData = {
            "database": self.dbName,
            "table": self.table,
            "adUnits": self._coneUnits,
            "sortDir": self._sortDir,
            "numRows": MAXROWS,
        }
        sendData.update(kwargs)

        # Specify columns, if we can
        if self._colsToGet is not None:
            sendData["cols"] = self._colsToGet
        elif self.defaultCols is not None:
            sendData["cols"] = self.defaultCols

        # And the sort Col:
        if self._sortCol is not None:
            sendData["sortBy"] = self._sortCol

        # Add filters
        if len(self._filters) > 0:
            sendData["constraints"] = []
            for f in self._filters:
                sendData["constraints"].append(f.data)

        # Add cone information
        if self._doConeSearch:
            if self._coneName is None:
                if (self._coneRA is None) or (self._coneDec is None):
                    raise RuntimeError("You have requested a cone search but without specifying name/position")
                sendData["searchRA"] = self._coneRA
                sendData["searchDec"] = self._coneDec
                # print("SETTING CONE BY POSITION")
            else:
                sendData["searchName"] = self._coneName
                # print("SETTING CONE BY NAME")

            sendData["searchRad"] = self._coneRadius
            if (self.raCol is not None) and (self.decCol is not None):
                sendData["RACol"] = self.raCol
                sendData["DecCol"] = self.decCol

        # Now do the actual work. Note - the server can only return so
        # many rows at once because of memory constraints. I will limit
        # the max in one go to base.MAXROWS. NB, if you try to return
        # more than this and the allocated memory is overrun you just
        # get a 500 error.

        fR = 0  # First row from this query
        if self._firstRow is not None:
            fR = int(self._firstRow)

        maxRows = 1e80  # i.e. BIG
        if self._maxRows is not None:
            maxRows = int(self._maxRows)
            # If we want < the maximum in one go, need to ammend numRows
            if maxRows < MAXROWS:
                sendData["numRows"] = maxRows

        # Create a local variable for the result for now, I don't want
        # to update self._results until the query has definitely succeeded.
        result = None
        done = False
        while not done:
            # Update the first row to get
            sendData["firstRow"] = fR

            if not self._silent:
                print(f"Calling DB look-up for rows {fR} -- {sendData['numRows'] + fR}")

            ret = base.submitAPICall(
                "queryDB",
                sendData,
                minKeys=["Results", "NumRows"],
                verbose=self._verbose,
            )
            if result is None:
                result = ret
            else:
                result["NumRows"] = result["NumRows"] + ret["NumRows"]
                result["Results"].extend(ret["Results"])

            # Are we done? If we did not get as many rows as was requested then we are
            if ret["NumRows"] < sendData["numRows"]:
                done = True
                if self._verbose:
                    print(f"Received {ret['NumRows']} rows / {sendData['numRows']} requested. Query complete.")
            # If we have now hit the user-set limit, then we are done:
            elif (self._maxRows is not None) and (result["NumRows"] >= self._maxRows):
                done = True
                if self._verbose:
                    print(f"{result['NumRows']} rows retrieved in total. Query complete.")
            else:
                # Increase the start row for the next call
                fR = fR + sendData["numRows"]
                # We may need to decrease the number of rows
                if (self._maxRows is not None) and (self._maxRows < result["NumRows"] + sendData["numRows"]):
                    sendData["numRows"] = self._maxRows - result["NumRows"]
                    if self._verbose:
                        print(f"Reducing the number of rows requested to {sendData['numRows']}.")
        # End of while not Done
        # We now should have our results. Maybe do one sanity check:
        if result["NumRows"] != len(result["Results"]):
            raise RuntimeError(f"Should have {result['NumRows']} rows, but have {len(result['Results'])}!")

        if (self._doConeSearch) and (not self.silent) and ("ResolvedInfo" in result):
            print(result["ResolvedInfo"])

        self._numRows = result["NumRows"]
        if (self._doConeSearch) and ("ResolvedInfo" in result):
            self._resolvedInfo = result["ResolvedInfo"]
            self._resolvedRA = result["ResolvedRA"]
            self._resolvedDec = result["ResolvedDec"]
        self._results = pd.DataFrame(result["Results"])

        # useAst = None
        # if base.HAS_ASTROPY:
        #     useAst = "_apy"
        if not self.silent:
            print(f"Received {self.numRows} rows.")
        self._manageResults("_s")
        self._locked = True

        # self._raw = result  # TEMPORARY LINE

    def _manageResults(self, ssuffix="_s"):
        """Give columns their correct types.

        This should be called on a pandas DataFrame created by the return
        from a query call. It will work through the metdata and convert
        numerical columns to numbers, and datetime columns to datetimes.
        It will also create sexagesimal versions of any coordinate columns.

        Parameters
        ----------

        ssuffix : str
            A suffix to add to column names for sexagesimal coordinates
            (default: '_s').

        """
        useAstropy = None
        if base.HAS_ASTROPY:
            useAstropy = "_apy"

        if self.verbose:
            print("Processing the returned self._results.")

        for c in self._results.columns:
            # Bit of a hack for angdist:
            action = 0  # 0 = Nothing, 1 = numeric, 2 = datetime, 3 = coordHr 4 = coordDeg
            # if self.verbose:
            #     print(f"Working on column {c}")
            if c == "_r":
                action = 1
            else:
                thisMD = self._metadata.loc[self._metadata["ColName"] == c]
                if len(thisMD) == 0:
                    raise ValueError(f"Column {c} is not in self._metadata, cannot parse self._results.")
                thisType = thisMD["Type"].tolist()[0]
                if (thisType == "NUM") or (thisType == "FLOAT") or (thisType == "INT"):
                    action = 1
                elif thisType == "UTC":
                    action = 2
                elif thisType == "COORDH":
                    action = 3
                elif thisType == "COORDD":
                    action = 4

                if ("IsObsCol" in thisMD) and (thisMD["IsObsCol"].iloc[0] == 1):
                    self._results[c] = self._results[c].apply(lambda a: f"{int(a):011d}")
                    action = -1

            if action == 1:
                if self.verbose:
                    print(f"Parsing column {c} as numeric")
                self._results[c] = pd.to_numeric(self._results[c])
            elif action == 2:
                if self.verbose:
                    print(f"Parsing column {c} as UTC self._results")
                self._results[c] = pd.to_datetime(self._results[c], yearfirst=True)
            elif action == 3:
                scol = f"{c}{ssuffix}"
                if self.verbose:
                    print(f"Parsing column {c} as coordinate, creating sexagesimal column `{scol}`")
                self._results[c] = pd.to_numeric(self._results[c])
                self._results[scol] = self._results[c].apply(lambda a: base.ra2sex(float(a)))
                if useAstropy is not None:
                    scol = f"{c}{useAstropy}"
                    if self.verbose:
                        print(f"Creating astropy.coordinates.Angle column `{scol}`")
                    self._results[scol] = self._results[c].apply(lambda a: base.makeAng(a))
            elif action == 4:
                scol = f"{c}{ssuffix}"
                if self.verbose:
                    print(f"Parsing column {c} as coordinate, creating sexagesimal column `{scol}`")
                self._results[c] = pd.to_numeric(self._results[c])
                self._results[scol] = self._results[c].apply(lambda a: base.dec2sex(float(a)))
                if useAstropy is not None:
                    scol = f"{c}{useAstropy}"
                    if self.verbose:
                        print(f"Creating astropy.coordinates.Angle column `{scol}`")
                    self._results[scol] = self._results[c].apply(lambda a: base.makeAng(a))

        # May also want to stringify the obsCol
        if (self.ObsIDAsString) and (self._obsCol is not None) and (self._obsCol in self._results.columns):
            c = self._obsCol
            self._results[c] = self._results[c].apply(lambda a: f"{a:011d}")

    # ---------------------------------------------------------------
    # Data retrieval

    def downloadObsData(self, subset=None, **kwargs):
        """Download data for the observations returned by the query.

        If the excuted query returned a column which includes
        observation identifiers, then this function will download the
        data for those columns. This function essentially wraps
        ukssdc.data.downloadObsData(), so for valid values of the
        ``**kwargs``, see the documentation for that function.

        The subset parameter is optional, and can be used to apply
        a filter to the results this query has obtained. The easiest way
        to generate this is using the pandas syntax for filtering on
        value.

        e.g. if your dataQuery object is called 'req' then:

        => req.downloadObsData(subset=req.results['xrt_expo_pc']<1000)

        would request a download of all observations that the request
        found, which had a value of less than 1000 in the 'xrt_expo_pc'
        column.

        Obviously, this function cannot be called before this request
        has been submitted and has completed; if you try, your computer
        will explode and your eyeballs will be eaten by ants (just
        kidding; you'll get a RuntimeError).

        Parameters
        ----------

        subset : pandas.Series, optional
            A pandas series defining a subset of rows to download.

        """

        if not self.haveResults:
            raise RuntimeError("This query has not been executed, cannot download!")

        if self._obsCol is None:
            raise RuntimeError("These is no column containing observation ID, cannot download.")

        if self._obsCol not in self.results.columns:
            raise RuntimeError(
                f"The column {self._obsCol} was not retrieved as part of your query. Cannot download. "
                "You may need to unlock this object, add {self._obsCol} to those to retrieve, and repeat the query."
            )

        obslist = []
        if subset is not None:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            obslist = self._results.loc[subset][self._obsCol].tolist()
        else:
            obslist = self.results[self._obsCol].tolist()

        # # DEBUG LINE:
        # print(f"Downloading obs: {obslist}")
        # return

        dl.downloadObsData(obslist, silent=self.silent, verbose=self.verbose, **kwargs)

    # ---------------------------------------------------------------
    # Ancillary things

    def doCatSearch(
        self,
        pandas=True,
        subset=None,
        idCol="Default",
        reIndex=False,
        what="simbad",
        returnData=False,
        **kwargs,
    ):
        """Do an astroquery catalogue query around the results.

        This will look for ``RA`` and ``Decl`` in your results, and if
        found, carry out SIMBAD lookup around those results.

        **This function requires the ``astroquery`` module.**

        The return type depends on a few things. The basic operation of
        this funciton is to return a list, with the results in the same
        order as the query results. This is far from ideal (as shown in
        a moment for subsets), so instead the ``idCol`` parameter can be
        specified, giving a column in the query results to use as an
        identifier for each row. Here, a ``dict`` is returned, with the
        keys being the values in this column. e.g. if ``idCol='ObsID'``
        then the dictionary keys will be the ``ObsID`` values for each
        entry in your query results.

        Some catalogues may have default values for ``idCol``, in which
        case these will be used unless otherwise specified. For example,
        for ``SXPS`` queries on the source catalogue, the ``LSXPS_ID``
        column is used by default. You can still override this by
        specifying the ``idCol`` parameter in this function call. If you
        want to force the return to be a list, i.e. not to even use the
        built-in default, then set ``idCol=None`` in the function call.

        If the ``subset`` parameter is specified then only the rows
        matching your subset definition will be queries. Note that in
        this case, if ``idCol``is ``None`` then the index in the list
        returned by this function will match the index in the
        ``self.results`` object by default, with missing indexes having
        values of ``None``. If ``reIndex=True`` then instead the results
        array will have indices 0..n, where n is the number of entries
        in the subset.

        Parameters
        ----------

        what : str, optional
            Which catalogue to search (default: "SIMBAD")

        pandas : bool, optional
            Whether to convert the results into a pandas
            ``DataFrame``. If not, an ``astropy.Table`` is returned
            (default: ``True``).

        subset : pandas.Series, optional

            A pandas series defining a subset of results to query.

        idCol : str, optional
            A column in the results set to use as in index for the
            returned data.

        reIndex : bool, optional
            Whether to reindex the results to 0-n, instead of matching
            the array indices from the original results (only affects)
            things if ``subset`` is supplied.

        **kwargs : dict
            Arguments to pass to astroquery.

        Returns:
        --------

        list or dict:
            The results from your query.

        """
        if self._results is None:
            raise RuntimeError("No results have been obtained yet!")

        if not base.HAS_ASTROQ:
            raise RuntimeError("astroquery does not appear to be installed.")

        allowedCats = ("simbad", "vizier")
        what = what.lower()
        if what not in allowedCats:
            raise RuntimeError(f"`{what}` is not a valid catalogue.")

        raCol = None
        decCol = None
        raCols = ("RA", "ra")
        decCols = ("Decl", "decl")

        for r in raCols:
            if r in self.results.columns:
                raCol = r
                break
        for d in decCols:
            if d in self.results.columns:
                decCol = d
                break

        if raCol is None:
            raise RuntimeError("Cannot find an RA column.")
        if decCol is None:
            raise RuntimeError("Cannot find a Dec column.")

        # So, handle the subset, and create results as list or dict based
        # on idCol.
        # Iterate and run the query.
        # Convert result to pandas if needed
        # Either push onto list, or dict.
        # Return. Do NOT store internally
        import warnings

        warnings.filterwarnings("ignore")

        entries = []
        if subset is None:
            entries = self.results
        else:
            if not isinstance(subset, pd.core.series.Series):
                raise ValueError("Subset parameter must be a pandas series")
            entries = self._results.loc[subset]

        if idCol == "Default":
            if self._defaultIDCol is not None:
                if self._defaultIDCol in entries.columns:
                    idCol = self._defaultIDCol
                elif not self.silent:
                    print(f"Column `{self._defaultIDCol}` was not in your query, so this will return a list.")

        # If idCol is still default, we can't get it, so make it None
        if idCol == "Default":
            idCol = None

        # Are we returning a dict or list?
        if idCol is not None:
            if idCol not in entries.columns:
                raise ValueError(f"`{idCol}` is not a column in your results.")
            ret = {}
        else:
            # If a list, then are we returning one per row in the results
            # DataFrame (default), or one per subset (if reIndex)?
            # We will create an list of Nones, so that I can just access
            # by index later on.
            if reIndex:
                t = len(entries)
            else:
                t = len(self.results)
            ret = [None] * t

        # Now do the actual queries.
        ix = 0
        for r in entries.iterrows():
            key = r[0]
            vals = r[1]
            if idCol is not None:
                key = vals[idCol]
            elif reIndex:
                key = ix
                ix = ix + 1
            ra = vals[raCol]
            dec = vals[decCol]
            pos = astropy.coordinates.SkyCoord(ra=ra, dec=dec, unit=(astropy.units.deg, astropy.units.deg), frame="fk5")
            if what == "vizier":
                r = aqv.Vizier.query_region(pos, **kwargs)
            elif what == "simbad":
                r = aqs.Simbad.query_region(pos, **kwargs)
            else:
                raise RuntimeError("Unknown query type")
            if pandas and isinstance(r, astropy.table.table.Table):
                r = r.to_pandas()
            ret[key] = r

        return ret

    def doSIMBADSearch(self, **kwargs):
        """Do a SIMBAD query around the results.

        This is just a wrapper to ``doCatSearch()`` so see the
        help for that for the arguments."""

        tmp = self.doCatSearch(what="simbad", **kwargs)

        if self._prodData["SIMBAD"] is None:
            self._prodData["SIMBAD"] = tmp
        else:
            self._prodData["SIMBAD"].update(tmp)

        if ("returnData" in kwargs) and kwargs["returnData"]:
            return tmp

    def doVizierSearch(self, pandas=True, **kwargs):
        """Do a SIMBAD query around the results.

        This is largely just a wrapper to ``doCatSearch()``
        but the conversion to Pandas happens in this function, so
        the ``pandas`` parameter is explicit here.

        """

        tmp = self.doCatSearch(what="vizier", pandas=False, **kwargs)
        # Vizier returns a TableList, which is a bit annoying.
        if pandas:
            if isinstance(tmp, dict):
                ix = tmp.keys()
            else:
                ix = range(len(tmp))

            for index in ix:
                i = tmp[index]
                if i is not None:
                    newOut = {}
                    for j in i.keys():
                        newOut[j] = i[j].to_pandas()
                    tmp[index] = newOut

        if self._prodData["Vizier"] is None:
            self._prodData["Vizier"] = tmp
        else:
            self._prodData["Vizier"].update(tmp)

        if ("returnData" in kwargs) and kwargs["returnData"]:
            return tmp
